package uibk.mtk.draw2d.axes2d;

import java.awt.BasicStroke;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.awt.geom.Point2D;

import uibk.mtk.draw2d.base.Scene2D;

/**
 * Ein "Tick" auf einer Achse
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
class Tick {

	/** Code f?r Position links von der Achse */
	static final int TL_LEFT = 0;

	/** Code f?r Position rechts von der Achse */
	static final int TL_RIGHT = 1;

	/** Code f?r Position ?ber der Achse */
	static final int TL_TOP = 2;

	/** Code f?r Position unter der Achse */
	static final int TL_BOTTOM = 3;

	/** Code das Beschriftung des Ticks zentiert wird */
	static final int LABEL_CENTER = 0;

	/** Code das Beschriftung des Ticks nach oben verschoben wird */
	static final int LABEL_MOVEUP = 1;

	/** Code das Beschriftung des Ticks nach oben verschoben wird */
	static final int LABEL_MOVEDOWN = 2;

	/** Code das Beschriftung des Ticks nach rechts verschoben wird */
	static final int LABEL_MOVERIGHT = 3;

	/** Code das Beschriftung des Ticks nach links verschoben wird */
	static final int LABEL_MOVELEFT = 4;

	/** gro?er/kleiner Tick */
	boolean bigTick;

	/** Postion des Ticks (Anfangspunkt) in Weltkoordinaten */
	Point2D.Double tickpos = new Point2D.Double();

	/** Postion der Beschriftung in Ger?tekoordinaten */
	Point labelpos = new Point();

	/** Postion der Beschriftung eine der LABEL Konstanten */
	int labelPos;

	/** Postion des Ticks eine der TL Konstanten */
	int pos;

	/** Beschriftung des Ticks */
	String label;

	/** L?nge eines gro?en Ticks in Pixeln */
	private static final int LENGTHBIGTICK = 6;

	/** L?nge eines kleinen Ticks in Pixeln */
	private static final int LENGTHSMALLTICK = 3;

	/** Achsenobjekt zu dem der Tick geh?rt */
	Axes2D axes;

	/** Abstand Tick und Beschriftung */
	private static final int SPACE = 10;

	Tick(int pos, double tickPosX, double tickPosY, boolean bigTick,
			String label, int labelPos, Axes2D axes) {
		this.pos = pos;
		tickpos.x = tickPosX;
		tickpos.y = tickPosY;
		this.bigTick = bigTick;
		this.labelPos = labelPos;
		this.label = label;
		this.axes = axes;
	}

	public void compute(Graphics2D g2, Scene2D coords2d) {

		if (!this.bigTick) {
			return;
		}

		int x = coords2d.xToPixel(tickpos.x);
		int y = coords2d.yToPixel(tickpos.y);

		Font oldFont = g2.getFont();
		if (axes.font != null)
			g2.setFont(axes.font);
		FontMetrics metrics = g2.getFontMetrics();
		int length = metrics.stringWidth(label);
		int height = metrics.getHeight();
		g2.setFont(oldFont);

		switch (this.pos) {
		case TL_LEFT:

			switch (labelPos) {
			case LABEL_MOVEUP:
				labelpos.y = y - LENGTHSMALLTICK;
				break;
			case LABEL_MOVEDOWN:
				labelpos.y = y + height / 2 + LENGTHSMALLTICK;
				break;
			default:
				labelpos.y = y + height / 4;
			}
			labelpos.x = x - length - SPACE;
			coords2d.requestSpaceLeft(labelpos.x);
			break;

		case TL_RIGHT:
			labelpos.x = x + SPACE;
			labelpos.y = y + height / 4;
			coords2d.requestSpaceRight(labelpos.x + length);

			break;
		case TL_TOP:
			labelpos.x = x - length / 2;
			labelpos.y = y - SPACE;
			coords2d.requestSpaceRight(labelpos.x + length);
			coords2d.requestSpaceLeft(labelpos.x);
			coords2d.requestSpaceTop(labelpos.y);
			break;
		case TL_BOTTOM:

			switch (labelPos) {
			case LABEL_MOVERIGHT:
				labelpos.x = x + LENGTHSMALLTICK;
				break;
			case LABEL_MOVELEFT:
				labelpos.x = x - length - LENGTHSMALLTICK;
				break;
			default:
				labelpos.x = x - length / 2;
			}
			labelpos.y = y + SPACE + height / 2;
			coords2d.requestSpaceRight(labelpos.x + length);
			coords2d.requestSpaceLeft(labelpos.x);
			coords2d.requestSpaceBottom(labelpos.y);

			break;
		}

	}

	public void drawTick(Graphics2D g2, Scene2D coords2d) {
		Stroke oldStroke = g2.getStroke();
		g2.setStroke(new BasicStroke((float) 0.5));
		int length = this.bigTick ? LENGTHBIGTICK : LENGTHSMALLTICK;

		int x = coords2d.xToPixel(tickpos.x);
		int y = coords2d.yToPixel(tickpos.y);

		switch (this.pos) {
		case TL_LEFT:
			g2.drawLine(x - length, y, x, y);
			break;
		case TL_RIGHT:
			g2.drawLine(x + length, y, x, y);
			break;
		case TL_TOP:
			g2.drawLine(x, y - length, x, y);
			break;
		case TL_BOTTOM:
			g2.drawLine(x, y + length, x, y);
			break;
		default:
			break;

		}
		g2.setStroke(oldStroke);
	}

	public void drawLabel(Graphics2D g2) {

		if (!this.bigTick) {
			return;
		}

		Font oldFont = g2.getFont();
		if (axes.font != null)
			g2.setFont(axes.font);
		if (label != null) {
			g2.drawString(label, labelpos.x, labelpos.y);

		}

		g2.setFont(oldFont);
	}

	/** Zeichnet den Tick. */

	public void draw(Graphics2D g2, Scene2D coords2d, boolean drawLabels) {
		this.drawLabel(g2);
		if (drawLabels) {
			this.drawTick(g2, coords2d);
		}
	}

}