package uibk.mtk.draw2d.objects;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.image.BufferedImage;
import java.util.Vector;

import uibk.mtk.draw2d.base.Drawable2D;
import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.util.GeomUtil;

/**
 * Ein zweidimensionaler Graph. Der Graph ist durch Punkte gegeben die durch
 * Linien in w?hlbarer Farbe und w?hlbareb Linienstil verbunden werden. (Falls
 * aktiviert werden die Linien am Rand des Koordinatensystems geclippt, wird
 * z.B. verwendet beim Zeichnen des Phasendiagramms im Applet ?ber DGLs)
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 * 
 */

public class Graph2D extends Drawable2D {

	/** Linienstile zum Zeichnen des Graphen */
	protected Stroke stroke;

	/** Punkte des Graphen (Elemente m?ssen vom Typ Punkt2D sein) */
	protected Vector points = new Vector();

	/** minimalen(maximalen Koordinaten des Graphen */
	protected CoordinateRect2D limits = new CoordinateRect2D();

	protected boolean limitsdirty = true;

	/** erzeugt einen Graphen in der Farbe rot mit einfachen Linienstil */
	public Graph2D() {
		super.color = Color.red;
		stroke = new BasicStroke();
	}

	/** setzt den Linienstil, in dem der Graph gezeichnet werden soll */
	public void setStroke(Stroke stroke) {
		this.stroke = stroke;
	}

	/** gibt den Linienstil, in dem der Graph gezeichnet werden soll zurck */
	public Stroke getStroke() {
		return stroke;
	}

	/**
	 * setzt die Punkte des Graphen, die minimalen Komponenten werden neu
	 * berechnet
	 */
	public void setPoints(Vector points) {
		this.points = points;
		limitsdirty = true;
	}

	/** f?gt einen Punkt zum Graphen hinzu */
	public void addPoint(Punkt2D p) {
		points.add(p);
		limitsdirty = true;
	}

	/** f?gt einen Punkt zum Graphen hinzu */
	public void addPoint(double x, double y) {
		points.add(new Punkt2D(x, y));
		limitsdirty = true;
	}

	/** Berechnet die Grenzen des Graphen neu */
	protected void recalcLimits() {

		if (points != null && points.size() > 0) {

			limits = GeomUtil.getLimits2D(points);
		} else {
			limits = new CoordinateRect2D(-1, 1, -1, 1);

		}
	}

	/**
	 * l?scht alle Punkte des Graphen. Minimale Koordinaten werden auf
	 * Standardwerte gesetzt
	 */
	public void clearpoints() {
		points.clear();
		limits = new CoordinateRect2D(-1, 1, -1, 1);

	}

	/** Lscht den ersten Punkt des Graphen */
	public void clearFirstPoint() {
		if (points != null)
			points.remove(0);
	}

	/** Gibt den ersten Punkt des Graphen zurck */
	public Punkt2D getFirstPoint() {
		if (points != null)
			return (Punkt2D) points.firstElement();
		else
			return null;
	}

	/** Gibt den letzen Punkt des Graphen zurck */
	public Punkt2D getLastPoint() {
		if (points != null && points.size()>0)
			return (Punkt2D) points.lastElement();
		else
			return null;
	}

	/** gibt die Punkte des Graphen zur?ck */
	public Vector getPoints() {
		return points;
	}

	/** gibt die Anzahl der Punkte des Graphen zur?ck */
	public int getNumberPoints() {
		return points.size();
	}

	/** gibt die Grenzen des Graphen zur?ck */
	public CoordinateRect2D getLimits() {

		if (limitsdirty) {
			recalcLimits();
		}

		return (CoordinateRect2D) limits.clone();

	}

	/**
	 * Zeichnet den Graphen, die Linien, aus denen Graph besteht werden geclippt
	 * falls clipping=true
	 */
	public void draw(BufferedImage framebuffer, Graphics2D g2) {
		Stroke oldStroke = g2.getStroke();
		g2.setStroke(stroke);
		g2.setColor(color);

		for (int i = 0; i < points.size() - 1; i++) {
			Punkt2D start = (Punkt2D) points.elementAt(i);
			Punkt2D end = (Punkt2D) points.elementAt(i + 1);
			if (start == null || end == null)
				continue;
			int startx = scene2d.xToPixel(start.x);
			int starty = scene2d.yToPixel(start.y);
			int endx = scene2d.xToPixel(end.x);
			int endy = scene2d.yToPixel(end.y);

			g2.setColor(color);
			g2.drawLine(startx, starty, endx, endy);

		}

		g2.setStroke(oldStroke);
	}

}