package uibk.mtk.draw2d.objects;

import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.util.Vector;

import uibk.mtk.draw2d.base.Drawable2D;
import uibk.mtk.draw2d.base.MathPanel2D;
import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.lang.ToolTipable;

/**
 * Ein zweidimensionaler Graph. Der Graph besteht aus <CODE>MathPoint2D</CODE>
 * Objekten die wahlweise durch Linien verbunden werden k?nnen. Die Punkte des
 * Graphen zeigen ihren Wert als Tooltip an.
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 *  
 */
public class PointPlot2D extends Graph2D implements ToolTipable {

	/**
	 * Farbe f?r die Linien (das von Drawable geerbte <CODE>color</CODE> wird
	 * f?r die Punkte verwendet)
	 */
	protected Color linecolor;

	/** Radius zum Zeichnen der Punkte */
	protected int pointradius;

	/** sind Tooltips aktiviert */
	protected boolean tooltipenabled;

	/** Punkte des Graphen (sind vom Typ MathPoint2D sein) */
	protected Vector points = new Vector();
	
	/** Linienstrke der Sulen */
	protected Stroke strokecolumns;

	/** gibt an ob die Punkte durch Linien verbunden werden sollen */
	boolean connectpoints = false;
	
	/** gibt an ob die Punkte durch Sulen dargestellt werden sollen*/
	boolean columns = false;


	/** setzt fest ob die Punkte durch Linien verbunden werden soll */
	public void connectPoints(boolean enable) {
		this.connectpoints = enable;
	}
	
	/** setzt fest die Punkte durch Sulen dargestellt werden sollen*/
	public void columns(boolean enable) {
		this.columns = enable;
		this.pointradius=0;
	}

	/** setzt den Radius zum Zeichnen der Punkte */
	public void setPointRadius(int pointradius) {
		this.pointradius = pointradius;
	}

	/** setzt die Farbe f?r die Linien */
	public void setLineColor(Color linecolor) {
		this.linecolor = linecolor;
	}

	/** setzt die Farbe f?r die Punkte */
	public void setColor(Color pointcolor) {
		super.setColor(pointcolor);
		for (int i = 0; i < points.size(); i++)
			((MathPoint2D) points.elementAt(i)).setColor(pointcolor);
	}

	/** ist Maus ?ber dem Tooltip-Bereich eines Punktes */
	public boolean tooltipcontains(int x, int y) {
		//alle Punkte pr?fen
		int ct = points.size();
		for (int i = 0; i < ct; i++) {
			//wir haben hier nur Punkte drin
			MathPoint2D p = (MathPoint2D) points.elementAt(i);
			if (p.tooltipcontains(x, y)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * gibt den Tooltip des Punktes ?ber den die Maus ist zur?ck (z.B.
	 * a(2)=21.123 bei 1. Punkt und letzten Punkt zus?tzlich die Formel der
	 * Folge)
	 */
	public String getToolTipText(MouseEvent ev) {

		int size = points.size();
		for (int i = 0; i < size; i++) {

			MathPoint2D p = (MathPoint2D) points.elementAt(i);

			if (p.tooltipcontains(ev.getX(), ev.getY())) {
				return "(" //$NON-NLS-1$
						+ p.getX() + "," + p.getY() + ")";} //$NON-NLS-1$ //$NON-NLS-2$
		}
		return null;

	}

	/** setzt den Tooltiptext (wird nicht verwendet) */
	public void setToolTipText(String tooltip) {
	}

	/** gibt zur?ck ob Tooltips aktiviert sind */
	public boolean isToolTipEnabled() {
		return this.tooltipenabled;
	}

	/** (de)aktiviert die Tooltips aktiviert sind */
	public void enableToolTip(boolean enabled) {
		this.tooltipenabled = enabled;
	}

	/** setzen des MathPanel f?r alle Punkte */
	public void setMathPanel2D(MathPanel2D panel) {
		super.setMathPanel2D(panel);
		int ct = points.size();
		for (int i = 0; i < ct; i++) {
			((Drawable2D) points.elementAt(i)).setMathPanel2D(panel);
		}
	}

	/**
	 * Erzeugt einen Graph fr Folgen mit angeschalteten Tooltips, Punkten in
	 * rot mit Radius 2 und Linien in blau
	 */
	public PointPlot2D() {
		linecolor = Color.blue;
		color = Color.red;
		pointradius = 2;
		stroke = new BasicStroke(1);
		strokecolumns = new BasicStroke(4);
		tooltipenabled = true;
	}

	/** Setzt die Punkte f?r den Graphen */
	public void setPoints(double[] x, double[] y) {
		points.clear();

		for (int i = 0; i < x.length; i++) {

			addMathPoint(x[i], y[i]);

		}
		limitsdirty = true;
	}

	/** f?gt einen Punkt zum Graphen hinzu */
	private void addMathPoint(double x, double y) {
		MathPoint2D mp = new MathPoint2D(x, y);
		mp.setMathPanel2D(this.panel);

		mp.setRadius(pointradius);
		mp.setColor(color);
		points.add(mp);
	}

	/** Zeichnen des Graphen */
	public void draw(BufferedImage framebuffer, Graphics2D g2) {

		int size = points.size();

		for (int i = 0; i < size; i++) {
			Drawable2D d = (Drawable2D) points.elementAt(i);

			if (d.isVisible()) {
				if (i < size - 1) {
					MathPoint2D start = (MathPoint2D) points.elementAt(i);
					MathPoint2D end = (MathPoint2D) points.elementAt(i + 1);
					g2.setColor(linecolor);
					g2.setStroke(stroke);

					if (connectpoints) {
						g2.drawLine(scene2d.xToPixel(start.getX()), scene2d
								.yToPixel(start.getY()), scene2d.xToPixel(end
								.getX()), scene2d.yToPixel(end.getY()));
					}
					if (columns) {
						g2.setStroke(strokecolumns);
						g2.drawLine(scene2d.xToPixel(start.getX()), scene2d
								.yToPixel(start.getY()), scene2d.xToPixel(start
								.getX()),scene2d.yToPixel(0));
					}

				}
				d.draw(framebuffer, g2);

			}

		}
		MathPoint2D point = (MathPoint2D) points.elementAt(size-1);
		if (columns) {
			g2.setStroke(strokecolumns);
			g2.drawLine(scene2d.xToPixel(point.getX()), scene2d
					.yToPixel(point.getY()), scene2d.xToPixel(point
					.getX()),scene2d.yToPixel(0));
		}

	}

	/** setzt die Punkte des Graphen */
	public void setPoints(Vector points) {
		this.points = points;
		limitsdirty = true;
	}

	/** f?gt einen Punkt zum Graphen hinzu */
	public void addPoint(Punkt2D p) {
		addMathPoint(p.x, p.y);
		limitsdirty = true;
	}

	/** f?gt einen Punkt zum Graphen hinzu */
	public void addPoint(double x, double y) {
		points.add(new Punkt2D(x, y));
		limitsdirty = true;
	}

	/** Berechnet die Grenzen des Graphen neu */
	protected void recalcLimits() {

		if (points != null && points.size() > 0) {

			limits = getLimits(points);
		} else {
			limits = new CoordinateRect2D(-1, 1, -1, 1);

		}
	}

	/**
	 * Berechnet die maximalen Koordinaten eines Vektors aus Punkt2D Elementen
	 *  
	 */
	public static CoordinateRect2D getLimits(Vector points) {

		double maximumx = Double.NEGATIVE_INFINITY;
		double maximumy = Double.NEGATIVE_INFINITY;
		double minimumx = Double.POSITIVE_INFINITY;
		double minimumy = Double.POSITIVE_INFINITY;

		for (int i = 0; i < points.size(); i++) {
			if (points.elementAt(i) != null
					&& ((MathPoint2D) points.elementAt(i)).x > maximumx) {
				maximumx = ((MathPoint2D) points.elementAt(i)).x;
			}
		}

		for (int i = 0; i < points.size(); i++) {

			if (points.elementAt(i) != null
					&& ((MathPoint2D) points.elementAt(i)).y > maximumy) {
				maximumy = ((MathPoint2D) points.elementAt(i)).y;
			}
		}

		for (int i = 0; i < points.size(); i++) {
			if (points.elementAt(i) != null
					&& ((MathPoint2D) points.elementAt(i)).x < minimumx) {
				minimumx = ((MathPoint2D) points.elementAt(i)).x;
			}
		}

		for (int i = 0; i < points.size(); i++) {
			if (points.elementAt(i) != null
					&& ((MathPoint2D) points.elementAt(i)).y < minimumy) {
				minimumy = ((MathPoint2D) points.elementAt(i)).y;
			}
		}

		return new CoordinateRect2D(minimumx, maximumx, minimumy, maximumy);
	}

}