package uibk.mtk.draw2d.util;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.lang.Math;

import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.geom.geom2d.Vector2D;
import uibk.mtk.util.Grid2D;
import uibk.mtk.draw2d.objects.DragablePoint2D;
;;

/**
 * Utility-Klasse die Funktionen zum Zeichnen einiger Objekte zur Verf?gung
 * stellt.
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public final class MiscDrawing2D {

	private MiscDrawing2D() {
	}

	/**
	 * zeichnet einen Pfeil der durch Kopf und Fu?punkt gegeben ist.
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param base
	 *            Fu?punkt des Pfeiles
	 * @param head
	 *            Kopfpunkt des Pfeiles
	 * @param lengtharrowhead
	 *            L?nge der Pfeilspitze
	 * @param offsetwings
	 *            Normalabstand der "Fl?gel"
	 */
	public static void drawArrow(Graphics2D g2, Punkt2D base, Punkt2D head,
			int lengtharrowhead, int offsetwings) {

		if (base.equals(head) || !base.isReal() || !head.isReal())
			return;

		Vector2D dir = base.sub(head).toVector();
		dir.normalize();
		if (Double.isNaN(dir.x) || Double.isInfinite(dir.x)
				|| Double.isNaN(dir.y) || Double.isInfinite(dir.y)
				|| (dir.x == 0 && dir.y == 0)) {
			return;
		}

		Punkt2D n = dir.normaldir();
		dir.scaleself(lengtharrowhead);
		n.scaleself(offsetwings);

		Punkt2D s = head.add(dir);
		Punkt2D left = s.add(n);
		Punkt2D right = s.add(Punkt2D.inverse(n));

		g2.drawLine((int) base.x, (int) base.y, (int) head.x, (int) head.y);
		g2.drawLine((int) left.x, (int) left.y, (int) head.x, (int) head.y);
		g2.drawLine((int) right.x, (int) right.y, (int) head.x, (int) head.y);
	}

	/**
	 * Zeichnet einen gef?llten Kreis mit gegebenen Radius und Mittelpunkt
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param x
	 *            x-Koordinate des Mittelpunkt des Kreises
	 * @param y
	 *            y-Koordinate des Mittelpunkt des Kreises
	 * 
	 * 
	 * @param radius
	 *            Radius des Kreises
	 */
	public static void drawCircle(Graphics2D g2, int x, int y, int radius) {
		g2.drawOval(x - radius, y - radius, 2 * radius + 1, 2 * radius + 1);
	}

	/**
	 * Zeichnet einen gef?llten Kreis mit gegebenen Radius und Mittelpunkt
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param x
	 *            x-Koordinate des Mittelpunkts
	 * @param y
	 *            y-Koordinate des Mittelpunkts
	 * @param radius
	 *            Radius des Kreises
	 */
	public static void fillCircle(Graphics2D g2, int x, int y, int radius) {
		g2.fillOval(x - radius, y - radius, 2 * radius + 1, 2 * radius + 1);
	}

	/**
	 * Zeichnet ein rechteckiges Gitter
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param xmin
	 *            minimale x-Koordinate des Gitters
	 * @param xmax
	 *            maximale x-Koordinate des Gitters
	 * @param ymin
	 *            minimale y-Koordinate des Gitters
	 * @param ymax
	 *            maximale y-Koordinate des Gitters
	 * @param nx
	 *            Anzahl der Gitterpunkte in x-Richtung
	 * @param ny
	 *            Anzahl der Gitterpunkte in y-Richtung
	 */
	public static void drawgrid(Graphics2D g2, int xmin, int xmax, int ymin,
			int ymax, int nx, int ny) {

		//Gitterpunkte erzeugen
		Punkt2D[][] grid = Grid2D.createGrid2D(xmin, xmax, ymin, ymax, nx, ny);

		//parallel y-Achse
		for (int i = 0; i < nx; i++)
			for (int j = 0; j < ny - 1; j++) {
				Punkt2D r = grid[i][j];
				Punkt2D s = grid[i][j + 1];
				g2.drawLine((int) r.x, (int) r.y, (int) s.x, (int) s.y);
			}
		//parallel y-Achse
		for (int i = 0; i < nx - 1; i++)
			for (int j = 0; j < ny; j++) {
				Punkt2D r = grid[i][j];
				Punkt2D s = grid[i + 1][j];
				g2.drawLine((int) r.x, (int) r.y, (int) s.x, (int) s.y);
			}
	}
	/**
	 * Zeichnet ein rechteckiges Gitter
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param xmin
	 *            minimale x-Koordinate des Gitters
	 * @param xmax
	 *            maximale x-Koordinate des Gitters
	 * @param ymin
	 *            minimale y-Koordinate des Gitters
	 * @param ymax
	 *            maximale y-Koordinate des Gitters
	 * @param nx
	 *            Anzahl der Gitterpunkte in x-Richtung
	 * @param ny
	 *            Anzahl der Gitterpunkte in y-Richtung
	 * @param gridpoints
	 *            Speichert Gitterpunkte
	 */
	public static void drawgriddrag(Graphics2D g2, int xmin, int xmax, int ymin,
			int ymax, int nx, int ny) {

		//Gitterpunkte erzeugen
		Punkt2D[][] grid = Grid2D.createGrid2D(xmin, xmax, ymin, ymax, nx, ny);

		Punkt2D[] anfang = new Punkt2D[2];
		anfang[0] = grid[0][0];
		anfang[1] = grid[nx-1][0];
		Punkt2D[] ende= new Punkt2D[2];
		ende[0] = grid[0][ny-1];
		ende[1] = grid[nx-1][ny-1];
		int m = Math.max(ny, nx);
		DragablePoint2D[][] verschieb = new DragablePoint2D[2][m-2];
		for (int i = 0; i < nx-2; i++) {
			verschieb[0][i] = new DragablePoint2D(grid[i+1][0].x,grid[i+1][0].y);
		}
		for (int i = 0; i < ny-2; i++) {
			verschieb[1][i] = new DragablePoint2D(grid[0][i+1].x,grid[0][i+1].y);
		}
		
		g2.drawLine((int) anfang[0].x,(int) anfang[0].y,(int) anfang[0].x,(int) ende[0].y);
		g2.drawLine((int) anfang[1].x,(int) anfang[1].y,(int) anfang[1].x,(int) ende[0].y);
		g2.drawLine((int) ende[0].x,(int) ende[0].y,(int) anfang[0].x,(int) ende[0].y);
		g2.drawLine((int) ende[1].x,(int) ende[1].y,(int) anfang[0].x,(int) ende[1].y);
		
		
		//parallel y-Achse

		for (int j = 0; j < nx - 2; j++) {
				DragablePoint2D r = verschieb[0][j];
				g2.drawLine((int) r.getX(), (int) r.getY(), (int) r.getX(), (int) ende[0].y);
		}
		for (int j = 0; j < ny - 2; j++) {
			DragablePoint2D r = verschieb[1][j];
			g2.drawLine((int) r.getX(), (int) r.getY(), (int) anfang[1].x, (int)r.getY());
		}

	}

	/**
	 * Zeichnet einen "Stern" mit gegebenen Radius und Mittelpunkt
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param p
	 *            Mittelpunkt des Kreises
	 * @param radius
	 *            Radius des Kreises
	 */
	public static void drawStar(Graphics2D g2, Point p, int radius) {
		int rhalf = radius / 2;

		g2.drawLine(p.x - rhalf, p.y, p.x + rhalf, p.y);
		g2.drawLine(p.x, p.y - rhalf, p.x, p.y + rhalf);

		g2.drawLine(p.x - rhalf, p.y - rhalf, p.x + rhalf, p.y + rhalf);
		g2.drawLine(p.x - rhalf, p.y + rhalf, p.x + rhalf, p.y - rhalf);

	}

	/**
	 * Zeichnet ein Label mit angegebner Farbe und Beschriftung
	 * 
	 * @param g2
	 *            Graphikkontext
	 * @param pos
	 *            Position des Labels (linke untere Ecke)
	 * @param label
	 *            Beschriftung
	 * @param background
	 *            Hintergrundfarbe
	 * @param fontcolor
	 *            Farbe f?r die Schrift
	 */
	public static void drawLabel(Graphics2D g2, Point pos, String label,
			Color background, Color fontcolor) {
		Color old = g2.getColor();
		FontMetrics metrics = g2.getFontMetrics();

		Rectangle2D rect = metrics.getStringBounds(label, g2);

		int w = (int) rect.getWidth() + 4;
		int h = (int) rect.getHeight();
		int x = pos.x - 2;
		int y = pos.y - (int) rect.getHeight() + 2;
		g2.setColor(background);
		g2.fillRect(x, y, w, h);
		g2.setColor(Color.black);
		g2.drawRect(x, y, w, h);
		g2.setColor(fontcolor);
		g2.drawString(label, pos.x, pos.y);
		g2.setColor(old);

	}

}