package uibk.mtk.draw3d.objects.surface3d;

import java.util.Vector;

import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.lang.ComputationException;
import uibk.mtk.lang.Messages;
import uibk.mtk.math.SingularityChecker;
import uibk.mtk.math.functions.Function1D;
import uibk.mtk.math.functions.Function2D;
import uibk.mtk.util.Grid2D;

/**
 * Klasse zum Erzeugen eines Gitters mit einer Funktion von R^2->R
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Function2DMeshBuilder {

	/** Funktion von R^2->R f?r deren Graph ein Mesh erzeugt wird */
	Function2D f;

	/** minimaler Wert der Funktion R^2->R */
	private double zmin;

	/** maximaler Wert der Funktion R^2->R */
	private double zmax;

	/**
	 * Vektor der f?r das Mesh verwendeten Vertices. Falls das Mesh ohne Test
	 * auf Singularit?ten erstellt wird sind die Elemente des Vektors genau die
	 * Elemente des Feldes vertices.
	 */
	private Vector mesh_vertices;

	/**
	 * Hilfsarray das festh?lt ob ein Vertex schon zu den Vertices f?r das Mesh
	 * hinzugef?gt wurde
	 */
	private boolean[] added;

	/**
	 * Array f?r die durch die Funktion und das Gitter definierten Vertices
	 *  
	 */
	Vertex3D[] vertices;

	/** Name des Resourcebundles */
	static final String BUNDLE_NAME = "uibk.applets.draw3d.objects.surface3d.messages";//$NON-NLS-1$

	public Function2DMeshBuilder(Function2D f) {
		this.f = f;
	}

	double dirx;

	double diry;

	double p1x;

	double p1y;

	Function1D g = new Function1D() {

		public double getValue(double t) {

			return f.getValue(p1x + dirx * t, p1y + diry * t);
		}

		public double getDerivation(double x) throws Exception {
			throw new RuntimeException(""); //$NON-NLS-1$
		}
	};

	/**
	 * Einfache Triangulierung der Fl?che, ohne dem Versuch Singularit?ten zu
	 * erkennen
	 */
	private Mesh simpleTriangulization(Vertex3D[] vertices, int nx, int ny)
			throws Exception {
		Vector triangles = new Vector((nx - 1) * (ny - 1) * 2);
		Vector vert = new Vector(nx * ny);
		for (int i = 0; i < vertices.length; i++)
			if (!vertices[i].mc.isReal())
				throw new ComputationException(Messages.getString(BUNDLE_NAME,
						"Function2DMeshBuilder.1")); //$NON-NLS-1$

		for (int k = 0; k < nx - 1; k++)
			for (int i = 0; i < ny - 1; i++) {

				//links unten
				int p1 = k * nx + i;
				//rechts unten
				int p2 = (k + 1) * nx + i;
				//rechts oben
				int p3 = (k + 1) * nx + i + 1;
				//links oben
				int p4 = k * nx + i + 1;

				triangles.add(new Triangle(vertices[p1], vertices[p2],
						vertices[p4]));
				triangles.add(new Triangle(vertices[p3], vertices[p4],
						vertices[p2]));

			}
		if (triangles.size() == 0)
			throw new ComputationException(Messages.getString(BUNDLE_NAME,
					"Function2DMeshBuilder.2")); //$NON-NLS-1$

		for (int i = 0; i < vertices.length; i++)
			vert.add(vertices[i]);

		return new Mesh(triangles, vert);

	}

	SingularityChecker singfinder;

	/**
	 * Triangulierung der Fl?che, mit dem Versuch Singularit?ten zu erkennen,
	 * und entsprechnde Dreiecke wegzulassen
	 */
	private Mesh advancedTriangulization(Vertex3D[] vertices, int nx, int ny)
			throws Exception {
		Vector triangles = new Vector((nx - 1) * (ny - 1) * 2);

		mesh_vertices = new Vector(nx * ny);

		singfinder = new SingularityChecker(zmin, zmax);

		//dummies
		Punkt2D a1 = new Punkt2D();
		Punkt2D a2 = new Punkt2D();
		Punkt2D a3 = new Punkt2D();
		Punkt2D a4 = new Punkt2D();

		for (int k = 0; k < nx - 1; k++)
			for (int i = 0; i < ny - 1; i++) {

				//links unten
				int p1 = k * nx + i;
				//rechts unten
				int p2 = (k + 1) * nx + i;
				//rechts oben
				int p3 = (k + 1) * nx + i + 1;
				//links oben
				int p4 = k * nx + i + 1;

				/*
				 * Achtung: bei z.B. 1/z falls Imagin?rteil fix bleibt und wir
				 * mit Realteil ?ber die Singularit?t gehen wird realteil immer
				 * kleiner aber Imagin?rteil gr??er und man bemekert Sing. am
				 * Realteil nicht ! Deswegen alle Seiten des Dreickes abgehen
				 */
				a1.setLocation(vertices[p1].mc.x1, vertices[p1].mc.x2);
				a2.setLocation(vertices[p2].mc.x1, vertices[p2].mc.x2);
				a3.setLocation(vertices[p3].mc.x1, vertices[p3].mc.x2);
				a4.setLocation(vertices[p4].mc.x1, vertices[p4].mc.x2);

				//?berpr?fen ob zw a1 und a2, bzw a2 und a4 Singularit?t ist
				if (!hasSingularity(a1, a2) && !hasSingularity(a2, a4)
						&& !hasSingularity(a4, a1)) {
					add(p1);
					add(p2);
					add(p4);

					triangles.add(new Triangle(vertices[p1], vertices[p2],
							vertices[p4]));
				}

				//?berpr?fen ob zw. a3 und a2, bzw a2 und a4 Singularit?t ist
				if (!hasSingularity(a3, a2) && !hasSingularity(a2, a4)
						&& !hasSingularity(a4, a3)) {
					add(p3);
					add(p2);
					add(p4);

					triangles.add(new Triangle(vertices[p3], vertices[p4],
							vertices[p2]));
				}

			}

		if (triangles.size() == 0)
			throw new Exception(Messages.getString(BUNDLE_NAME,
					"Function2DMeshBuilder.3")); //$NON-NLS-1$
		return new Mesh(triangles, mesh_vertices);

	}

	/**
	 * Hinzuf?gen des i-ten Vertex (falls er nicht schon hinzugef?gt worden ist)
	 */
	private void add(int i) {
		if (!added[i]) {
			mesh_vertices.add(vertices[i]);
			added[i] = true;
		}
	}

	/**
	 * Erzeugen der Fl?che ?ber dem durch <CODE>rect</CODE> festgelegten Teil
	 * des Koordinatensystems.
	 * 
	 * @param rect
	 *            Rechteck ?ber dem die Fl?che gezeichnet wird
	 * @param nx
	 *            Anzahl Gitterpunkte in x-Richtung
	 * @param ny
	 *            Anzahl Gitterpunkte in y-Richtung
	 * @param checksing
	 *            auf Singularit?ten testen und diese Dreiecke aussortieren
	 * @return
	 * @throws Exception
	 *             bei ?berlauf etc.
	 */
	public Mesh build(CoordinateRect2D rect, int nx, int ny, boolean checksing)
			throws Exception {

		vertices = new Vertex3D[nx * ny];
		added = new boolean[nx * ny];
		//Gitter in der Ebene
		Punkt2D[][] grid = Grid2D.createGrid2D(rect.xmin, rect.xmax, rect.ymin,
				rect.ymax, nx, ny);

		double z = 0;

		//Berechnung der Funktionswerte ?ber dem Gitter
		zmin = Double.POSITIVE_INFINITY;
		zmax = Double.NEGATIVE_INFINITY;

		int c = 0;
		for (int i = 0; i < nx; i++)
			for (int j = 0; j < ny; j++) {

				z = f.getValue(grid[i][j].x, grid[i][j].y);

				if (!Double.isInfinite(z) && !Double.isNaN(z)) {
					if (z < zmin)
						zmin = z;
					if (z > zmax)
						zmax = z;
				}

				vertices[c] = new Vertex3D(grid[i][j].x, grid[i][j].y, z);
				c++;
			}

		//Triangulierung der Fl?che
		if (checksing)
			return advancedTriangulization(vertices, nx, ny);
		else
			return simpleTriangulization(vertices, nx, ny);

	}

	/**
	 * Testen ob auf der Strecke zw. den Punkten p1 und p2 eine Singularit?t
	 * auftritt
	 *  
	 */
	private boolean hasSingularity(final Punkt2D p1, final Punkt2D p2) {

		dirx = p2.x - p1.x;
		diry = p2.y - p1.y;
		p1x = p1.x;
		p1y = p1.y;

		return singfinder.hasSingularity(g, 0, 1);

	}

}