package uibk.mtk.draw3d.rasterizer;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.awt.image.BufferedImage;

import uibk.mtk.draw3d.objects.surface3d.Triangle;
import uibk.mtk.draw3d.objects.surface3d.Vertex3D;

/**
 * Rasterizer zum F?llen von Dreiecken mit flat oder gouraud-shading und
 * Zeichnen von Linien.
 * 
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Rasterizer {

	/** der zu verwendende Tiefenbuffer, in den gezeichnet wird */
	BufferedImage framebuffer;

	/** der zu verwendende Tiefenbuffer */
	double[][] zbuffer;

	/** Graphikkontext */
	Graphics2D g2;

	/** Konstante f?r flat-shading */
	public static final int FLAT = 0;

	/** Konstante f?r gouroud-shading */
	public static final int GOUROUD = 1;

	/** setzt die Art der Schattierung (flat,gouraud) fest */
	int shadingmode = GOUROUD;

	/** aktuell zu verwendende Farbe */
	int colorint;

	/** aktuell zu verwendende Farbe */
	Color color;

	/** Breite zum Zeichnen einer Linie */
	private int linewidth = 1;

	/** Renderoption: RENDER_HINT_SPEED/RENDER_HINT_QUALITY */
	int renderhint = RENDER_HINT_QUALITY;

	/** Code f?r schnelles Zeichnen */
	public static final int RENDER_HINT_SPEED = 0;

	/** Code f?r Zeichnen mit bester Qualit?t */
	public static final int RENDER_HINT_QUALITY = 1;

	/**
	 * legt fest ob z-Buffer verwendet werden soll. Falls ja wird ?ber den
	 * Tiefenbuffer entschieden ob Pixel gezeichnet wird und falls er gezeichnet
	 * wird, wird Tiefe des Pixels im zBuffer eingetragen. Falls z-Buffer nicht
	 * verwendet werden soll wird der Pixel immer gezeichnet, und seine Tiefe
	 * falls er n?her ist wie bisheriger Eintrag, in Tiefenbuffer geschrieben.
	 */
	private boolean usezbuffer = false;

	//REGISTER
	private int scanline;

	private double xleft, xright;

	private Farbe colorleft = new Farbe();

	private Farbe colorright = new Farbe();

	private Farbe currentcolor = new Farbe();

	//Offset der Farbe pro scanline links
	private Farbe dcolorL = new Farbe();

	//  Offset der Farbe pro scanline rechts
	private Farbe dcolorR = new Farbe();

	//Tiefe links,rechts
	private double depthR, depthL;

	//linke und rechte Grenze f?r fillline
	int a, b;

	//Farb links und rechts der scanline
	Farbe colorL, colorR;

	//offset f?r den Tiefenpuffer in der Scanline
	double dD;

	//Tiefe in der Scanline
	double depth;

	//Offset f?r Tiefenbuffer links,rechts
	double ddepthR, ddepthL;

	private Point top = new Point(), middle = new Point(),
			bottom = new Point();

	private Farbe colortop, colormiddle, colorbottom;

	private double depthtop, depthmiddle, depthbottom;

	/** Setzt die Schattierungsoption: FLAT oder GOUROUD */
	public void setShadingMode(int mode) {
		this.shadingmode = mode;
	}

	/** Gibt die Schattierungsoption zur?ck: FLAT oder GOUROUD */
	public int getShadingMode() {
		return shadingmode;
	}

	/**
	 * Setzt die Farbe die zum Zeichnen der Linien und beim FLAT-Shading
	 * verwendet wird
	 */
	public void setColor(Color c) {
		this.color = c;
		this.colorint = color.getRGB();
	}

	/** Setzt den RenderHint: RENDER_HINT_SPEED oder RENDER_HINT_QUALITY */
	public void setRenderingHint(int hint) {
		if (hint > 1 || hint < 0)
			throw new IllegalArgumentException("rendering hint not specified"); //$NON-NLS-1$

		renderhint = hint;
	}

	/** Gibt den RenderHint zur?ck: RENDER_HINT_SPEED oder RENDER_HINT_QUALITY */
	public int getRenderingHint() {

		return renderhint;
	}

	/** Setzt die Buffer */
	public void setBuffers(BufferedImage framebuffer, Graphics2D g2,
			double[][] zbuffer) {
		this.zbuffer = zbuffer;
		this.framebuffer = framebuffer;
		this.g2 = g2;

	}

	/**
	 * setzt fest ob der Tiefenbuffer zum Entscheiden ob Pixel gezeichnet wird
	 * oder nicht herangezogen wird
	 */
	public void enableZBuffer(boolean enable) {
		this.usezbuffer = enable;

	}

	/** Sortieren der Eckpunkte des Dreiecks */
	private void sortEdges(Triangle triangle) {

		top.x = triangle.p1.dc.x;
		top.y = triangle.p1.dc.y;
		middle.x = triangle.p2.dc.x;
		middle.y = triangle.p2.dc.y;
		bottom.x = triangle.p3.dc.x;
		bottom.y = triangle.p3.dc.y;

		colortop = new Farbe(triangle.p1.color);
		colormiddle = new Farbe(triangle.p2.color);
		colorbottom = new Farbe(triangle.p3.color);

		depthtop = triangle.p1.distance;
		depthmiddle = triangle.p2.distance;
		depthbottom = triangle.p3.distance;

		Point temp;
		Farbe tempC;
		double tempdepth;
		if (top.y > middle.y) {
			temp = middle;
			middle = top;
			top = temp;

			tempdepth = depthmiddle;
			depthmiddle = depthtop;
			depthtop = tempdepth;

			tempC = colormiddle;
			colormiddle = colortop;
			colortop = tempC;
		}
		if (middle.y > bottom.y) {
			temp = middle;
			middle = bottom;
			bottom = temp;

			tempdepth = depthmiddle;
			depthmiddle = depthbottom;
			depthbottom = tempdepth;

			tempC = colormiddle;
			colormiddle = colorbottom;
			colorbottom = tempC;
		}
		if (top.y > middle.y) {
			temp = middle;
			middle = top;
			top = temp;

			tempdepth = depthmiddle;
			depthmiddle = depthtop;
			depthtop = tempdepth;

			tempC = colormiddle;
			colormiddle = colortop;
			colortop = tempC;
		}

	}

	/**
	 * F?llt ein Dreieck (wie gerendert wird, wird durch shadingmode,usezbuffer
	 * und color festgelegt
	 * 
	 * @param tri
	 *            zu renderndes Dreieck
	 */
	public void renderTriangle(Triangle tri) {

		//setzen der Register
		sortEdges(tri);

		scanline = top.y >= 0 ? top.y : 0;

		//offsets
		double dxL, dxR;

		//speziallf?lle horizontale Linien
		if ((top.y == middle.y) && (bottom.y == top.y)) {
			xleft = Math.min(Math.min(top.x, middle.x), bottom.x);
			xright = Math.max(Math.max(top.x, middle.x), bottom.x);
			if (xleft == top.x) {
				colorleft.set(colortop);
				depthL = depthtop;
			}
			if (xleft == bottom.x) {
				colorleft.set(colorbottom);
				depthL = depthbottom;
			}
			if (xleft == middle.x) {
				colorleft.set(colormiddle);
				depthL = depthmiddle;
			}
			if (xright == top.x) {
				colorright.set(colortop);
				depthR = depthtop;
			}
			if (xright == bottom.x) {
				colorright.set(colorbottom);
				depthR = depthbottom;
			}
			if (xright == middle.x) {
				colorright.set(colormiddle);
				depthR = depthmiddle;
			}

			renderline();

			return;
		}

		//obere H?lfte f?llen, falls middle.y == top.y gibt es keine obere
		// H?lfte
		if (middle.y != top.y) {

			dxL = (double) (middle.x - top.x) / (middle.y - top.y);
			dxR = (double) (bottom.x - top.x) / (bottom.y - top.y);
			dcolorL = colormiddle.sub(colortop);
			dcolorL.scaleself((float) (1.0 / (middle.y - top.y)));
			dcolorR = colorbottom.sub(colortop);
			dcolorR.scaleself((float) (1.0 / (bottom.y - top.y)));

			xleft = top.x;
			xright = top.x;
			colorright.set(colortop);
			colorleft.set(colortop);

			depthR = depthtop;
			depthL = depthtop;

			ddepthL = (depthmiddle - depthtop) / (middle.y - top.y);
			ddepthR = (depthbottom - depthtop) / (bottom.y - top.y);

			while (scanline <= middle.y) {
				renderline();
				xleft += dxL;
				xright += dxR;
				colorright.addself(dcolorR);
				colorleft.addself(dcolorL);
				depthR += ddepthR;
				depthL += ddepthL;
				scanline++;

			}
		}
		//untere H?lfte f?llen
		if (middle.y != bottom.y) {
			scanline = middle.y;

			dxL = (double) (middle.x - bottom.x) / (middle.y - bottom.y);
			dxR = (double) (top.x - bottom.x) / (top.y - bottom.y);

			dcolorL = colormiddle.sub(colorbottom);
			dcolorL.scaleself((float) (1.0 / (middle.y - bottom.y)));
			dcolorR = colortop.sub(colorbottom);
			dcolorR.scaleself((float) (1.0 / (top.y - bottom.y)));

			xleft = middle.x;
			xright = bottom.x + (scanline - bottom.y) * dxR;

			colorleft.set(colormiddle);
			colorright = Farbe.interpolate(top.y, bottom.y, colortop,
					colorbottom, scanline);

			ddepthL = (depthmiddle - depthbottom) / (middle.y - bottom.y);
			ddepthR = (depthtop - depthbottom) / (top.y - bottom.y);
			depthL = depthmiddle;
			depthR = depthbottom + (scanline - bottom.y) * ddepthR;

			//Scanline verschieben,Tiefe,Grenzen,Farbe aktualisieren
			while (scanline <= bottom.y) {
				renderline();
				xleft += dxL;
				xright += dxR;
				colorright.addself(dcolorR);
				colorleft.addself(dcolorL);
				depthR += ddepthR;
				depthL += ddepthL;
				scanline++;

			}

		}
	}

	/** Rendert eine Linie des Dreiecks */
	private void renderline() {

		a = (int) xleft;
		b = (int) xright;

		if (scanline > framebuffer.getHeight() - 1 || scanline < 0)
			return;

		int temp;
		double tempdepth;
		colorL = new Farbe(colorleft);
		colorR = new Farbe(colorright);
		Farbe colortemp;
		if (a > b) {
			temp = a;
			a = b;
			b = temp;
			tempdepth = depthL;
			depthL = depthR;
			depthR = tempdepth;
			colortemp = colorL;
			colorL = colorR;
			colorR = colortemp;
		}

		//"Clippen"
		if (a < 0)
			a = 0;
		if (b >= framebuffer.getWidth())
			b = framebuffer.getWidth() - 1;

		//Offset f?r Tiefenpuffer berechnen
		dD = 0;
		depth = depthL;

		if (b - a != 0)
			dD = (depthR - depthL) / (b - a);

		if (shadingmode == FLAT) {

			renderlineFlat();
		} else {

			renderlineGouroud();
		}
	}

	/**
	 * Rendern einer Scanline mit Interpolation der Farben ohne Beachtung des
	 * ZBuffer
	 *  
	 */
	private void renderlineGouroud() {

		int x = a;

		currentcolor.set(colorL);

		Farbe dcolor = new Farbe(0, 0, 0);
		if (b - a != 0) {
			dcolor.set(colorR.sub(colorL));
			dcolor.scaleself((float) (1.0 / Math.abs(b - a)));
		}

		while (x <= b) {
			if (!usezbuffer || depth < zbuffer[x][scanline]) {
				if (depth < zbuffer[x][scanline])
					zbuffer[x][scanline] = depth;
				framebuffer.setRGB(x, scanline, currentcolor.toInteger());

			}
			depth += dD;
			currentcolor.addself(dcolor);
			/*
			 * c_red += c_off_red; c_green += c_off_green;
			 */
			x++;
		}
	}

	/**
	 * Rendern einer Scanline mit der aktuellen Farbe <CODE>color</CODE> mit
	 * Beachtung des ZBuffer zu beachten
	 *  
	 */
	private void renderlineFlat() {
		int x = a;

		while (x <= b) {

			if (!usezbuffer || depth < zbuffer[x][scanline]) {
				if (depth < zbuffer[x][scanline])
					zbuffer[x][scanline] = depth;
				framebuffer.setRGB(x, scanline, colorint);
			}
			x++;
			depth += dD;
		}

	}

	/**
	 * Zeichnet eine Linie mit Hilfe des Bresenham-Algorithmus Vorlage: Skript
	 * Computergraphik Olaf M?ller /Oliver Vornberger Uni Osnabr?ck
	 * 
	 * 
	 * @param pdistance
	 *            Tiefe des Punktes p1
	 * @param qdistance
	 *            Tiefe des Punktes p2
	 * 
	 *  
	 */
	private void plotline(int px, int py, int qx, int qy, double pdistance,
			double qdistance) {

		//Point p = new Point((int) p1.x, (int) p1.y);
		//Point q = new Point((int) p2.x, (int) p2.y);

		int error, delta, schwelle, dx, dy, incx, incy;

		//f?r z-Bufferberechnung
		double depth = pdistance;

		double n = Math.sqrt((px - qx) * (px - qx) + (py - qy) * (py - qy)) + 1;
		//offset zBuffer=
		double dd = (qdistance - pdistance) / n;

		int width = framebuffer.getWidth();
		int height = framebuffer.getHeight();

		dx = qx - px;
		dy = qy - py;

		if (dx > 0)
			incx = 1;
		else
			incx = -1;
		if (dy > 0)
			incy = 1;
		else
			incy = -1;

		if (Math.abs(dy) < Math.abs(dx)) { //flach nach oben oder flach nach
			// unten

			error = -Math.abs(dx);
			delta = 2 * Math.abs(dy);
			schwelle = 2 * error;

			while (px != qx) {

				//Interpolation der Tiefe im aktuellen Punkt
				if ((px >= 0 && px < width) && (py >= 0 && py < height)) {
					depth += dd;

					if (!usezbuffer || depth < zbuffer[px][py]) {

						setPixel(px, py);

					}

				}

				px += incx;
				error += delta;
				if (error > 0) {
					py += incy;
					error += schwelle;
				}
			}
		} else { //Steil nach oben/unten

			error = -Math.abs(dy);
			delta = 2 * Math.abs(dx);
			schwelle = 2 * error;
			while (py != qy) {

				depth += dd;
				if ((px >= 0 && px < width) && (py >= 0 && py < height)) {
					if (!usezbuffer || depth < zbuffer[px][py]) {
						setPixel(px, py);
					}
				}
				py += incy;
				error += delta;
				if (error > 0) {
					px += incx;
					error += schwelle;
				}
			}
		}

	}

	/** Setzt die Linienbreite */
	public void setLineWidth(int w) {
		this.linewidth = w;
	}

	/** Gibt die Linienbreite zur?ck */
	public int getLineWidth() {
		return linewidth;
	}

	/** Zeichnen einer Linie */
	public void drawline(int px, int py, int qx, int qy, double pdepth,
			double qdepth) {

		Vertex3D p = new Vertex3D();
		Vertex3D q = new Vertex3D();
		p.dc = new Point(px, py);
		q.dc = new Point(qx, qy);
		p.distance = pdepth;
		q.distance = qdepth;
		drawline(p, q);

	}

	/** Zeichnen einer Linie */
	public void drawline(Vertex3D p, Vertex3D q) {
		int px = p.dc.x;
		int py = p.dc.y;
		int qx = q.dc.x;
		int qy = q.dc.y;

		double pdistance = p.distance;
		double qdistance = q.distance;
		g2.setColor(color);
		Stroke oldstroke = g2.getStroke();
		g2.setStroke(new BasicStroke(linewidth));

		switch (renderhint) {

		case RENDER_HINT_SPEED:

			plotline(px, py, qx, qy, pdistance, qdistance);
			break;

		case RENDER_HINT_QUALITY:
			if (usezbuffer) {
				plotline(px, py, qx, qy, pdistance, qdistance);
			} else {
				g2.setColor(color);
				g2.drawLine(px, py, qx, qy);
			}

			break;
		default:
			break;
		}

		g2.setStroke(oldstroke);
	}

	/** Zeichnen eines Pixels */
	private void setPixel(int x, int y) {
		if (renderhint == RENDER_HINT_SPEED)
			framebuffer.setRGB(x, y, colorint);
		else

			g2.drawLine(x, y, x, y);
	}

}