package uibk.mtk.geom.geom3d;

/**
 * Ein Punkt im Raum gegeben durch x,y und z-Koordinate (und homogener
 * Koordinate, die zum Rechnen der Transformation in der Klasse Scene3D zum
 * Einsatz kommt)
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Punkt3D {

	/** "x"-Koordinate */
	public double x1;

	/** "y"-Koordinate */
	public double x2;

	/** "z"-Koordinate */
	public double x3;

	/** homogene Koordinate */
	public double h = 1;

	/** Erzeugt den Punkt mit den ?bergebenen Koordinaten */
	public Punkt3D(double x, double y, double z) {
		this.x1 = x;
		this.x2 = y;
		this.x3 = z;
	}

	/** Erzeugt den Punkt mit den Koordinaten (0,0,0) */
	public Punkt3D() {
	}

	/**
	 * Erzeugt den Punkt mit den Koordinaten des ?bergebenen Punktes (ohne der
	 * homogenen Koordinate
	 */
	public Punkt3D(Punkt3D p) {
		this.x1 = p.x1;
		this.x2 = p.x2;
		this.x3 = p.x3;
	}

	/** dividiert x,y und z Koordinate durch die homogene Koordinate */
	public void toCanonical() {
		x1 /= h;
		x2 /= h;
		x3 /= h;
	}

	/** Addition zweier Punkte */
	public Punkt3D add(Punkt3D b) {
		return new Punkt3D(x1 + b.x1, x2 + b.x2, x3 + b.x3);
	}

	/** Subtraktion zweier Punkte */
	public Punkt3D sub(Punkt3D v) {
		return new Punkt3D(x1 - v.x1, x2 - v.x2, x3 - v.x3);
	}

	/** additives Inverse des Punktes */
	public static Punkt3D inverse(Punkt3D a) {
		return new Punkt3D(-a.x1, -a.x2, -a.x3);
	}

	/** setzt die Koordinaten des Punktes */
	public void setLocation(Punkt3D p) {
		this.x1 = p.x1;
		this.x2 = p.x2;
		this.x3 = p.x3;

	}

	/** setzt die Koordinaten des Punktes */
	public void setLocation(double x1, double x2, double x3) {
		this.x1 = x1;
		this.x2 = x2;
		this.x3 = x3;
	}

	/** Skalierung der Koordinaten */
	public void scaleself(double s) {
		this.x1 *= s;
		this.x2 *= s;
		this.x3 *= s;
	}

	/** Skalierung der Koordinaten */
	public Punkt3D scale(double s) {
		Punkt3D b = new Punkt3D();
		b.x1 = x1 * s;
		b.x2 = x2 * s;
		b.x3 = x3 * s;
		return b;
	}

	/** Berechnet das Kreuzprodukt zweier (als Vektoren interpretierter) Punkte */
	public Vector3D crossProd(Punkt3D b) {
		Vector3D c = new Vector3D();
		c.x1 = x2 * b.x3 - x3 * b.x2;
		c.x2 = x3 * b.x1 - x1 * b.x3;
		c.x3 = x1 * b.x2 - x2 * b.x1;
		return c;
	}

	/** Berechnet das Skalarprodukt zwei Punkte */
	public double dotProd(Punkt3D a) {
		return a.x1 * x1 + a.x2 * x2 + a.x3 * x3;
	}

	/** berechnet die Norm des Punktes */
	public double norm() {
		return Math.sqrt(x1 * x1 + x2 * x2 + x3 * x3);
	}

	/** Berechnet das Quadrat der Norm des Punktes */
	public double quadnorm() {
		return x1 * x1 + x2 * x2 + x3 * x3;
	}

	/** skaliert die Komponenten des Punktes so, dass er die Norm 1 hat */
	public void normalize() {
		double n = Math.sqrt(x1 * x1 + x2 * x2 + x3 * x3);

		x1 /= n;
		x2 /= n;
		x3 /= n;
	}

	/** gibt eine Zeichenkette der Form (x1,x2,x3) zur?ck */
	public String toString() {
		return "(" + x1 + "," + x2 + "," + x3 + ")"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
	}

	/**
	 * gibt zur?ck ob die Koordinaten des Punktes endlich sind.
	 */
	public boolean isInfinite() {
		if (Double.isInfinite(x1) || Double.isInfinite(x2)
				|| Double.isInfinite(x3)) {
			return true;
		}

		return false;
	}

	/**
	 * Gibt <CODE>false</CODE> falls eine der Koordinaten <CODE>NaN</CODE>
	 * ist
	 */
	public boolean isNaN() {
		if (Double.isNaN(x1) || Double.isNaN(x2) || Double.isNaN(x3))
			return true;

		return false;
	}

	/**
	 * gibt zur?ck ob die Koordinaten reell sind, d.h. keine der Koordinaten NaN
	 * oder Infinite ist.
	 */
	public boolean isReal() {
		if (isNaN() || isInfinite())
			return false;

		return true;
	}

	/** Erstellt eine Kopie des Punktes */
	public Object clone() {
		return new Punkt3D(x1, x2, x3);
	}
}