package uibk.mtk.math;

import java.util.Hashtable;
import java.util.Vector;

import uibk.mtk.lang.ComputationException;
import uibk.mtk.lang.InputException;
import uibk.mtk.lang.Messages;
import uibk.mtk.parser.realparser.Expr;
import uibk.mtk.parser.realparser.Variable;

/**
 * Klasse zum Berechnen expliziter und rekursiver Folgen in den reellen Zahlen.
 * Erlaubt sind Formeln in denen als "Variablen" "n" und Ausdr?cke der form
 * a(n-i), i>0 vorkommen als z.B. n*a(n-1)+a(n-3)^2
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Sequence {

	/** Hashtable mit den Variablen */
	Hashtable variables = new Hashtable();

	/** die Startwerte f?r eine rekursive Folge */
	private double[] initial;

	/** Formel f?r die Folge */
	Expr formula;

	/** Startindex */
	int startindex;

	/** Endindex */
	int endindex;

	/** maximale Anzahl von Folgengliedern die berechnet werden soll */
	int maxlengthsequence = 1000;

	/** rekursive Folge */
	public static final int RECURSIVE = 0;

	/** explizite Folge */
	public static final int EXPLICIT = 1;

	/**
	 * speichert die Indices der rekursiven Teile der Formel im Falle
	 * a(n-1)+a(n-3) also z.B 1,3
	 */
	private Vector rekindices;

	/**
	 * speichert die Tiefe der Rekursion.Iim Falle a(n-1)+a(n-3) also z.B 3, bei
	 * einer expliziten Folge gleich 0. Ist gleich der Anzahl der Elemente im
	 * Vektor rekindices
	 */
	private int sequencedepth;

	/** Name des Resourcebundles */
	static final String BUNDLE_NAME = "uibk.mtk.math.messages";//$NON-NLS-1$

	/**
	 * Erzeugt eine Folge, die die maximal die Berechnung von 1000
	 * Folgengliedern erlaubt
	 */
	public Sequence(Hashtable variables, Expr formula, Vector rekindices,
			int depth) {
		this.variables = variables;
		this.formula = formula;
		this.sequencedepth = depth;
		this.rekindices = rekindices;

	}

	/**
	 * legt die maximale Anzahl der zu berechnenden Folgenglieder fest (falls
	 * endindex-startindex gr??er als dieser Wert, gibt es beim Parsen eine
	 * Exception
	 */
	public void setMaxSequenceLength(int maxlength) {
		this.maxlengthsequence = maxlength;
	}

	/** Setzt die Anfangswerte der Folge (falls rekursiv) */
	public void setInitialValues(double[] initialvalues) {
		this.initial = initialvalues;
	}

	/** gibt den Startindex zur?ck */
	public int getStartIndex() {
		return startindex;
	}

	/** gibt den Endindex zur?ck */
	public int getEndIndex() {
		return endindex;
	}

	/** ?berpr?fen on alle Eingabe korrekt sind */
	private void check() throws Exception {
		if (variables == null || formula == null)
			throw new InputException(Messages.getString(BUNDLE_NAME,
					"Sequence.0")); //$NON-NLS-1$

		if (this.endindex < 0)
			throw new InputException(Messages.getString(BUNDLE_NAME,
					"Sequence.1")); //$NON-NLS-1$

		if (this.startindex < 0)
			throw new InputException(Messages.getString(BUNDLE_NAME,
					"Sequence.2")); //$NON-NLS-1$

		if (this.endindex < this.startindex)
			throw new Exception(Messages.getString(BUNDLE_NAME, "Sequence.3")); //$NON-NLS-1$

		if (getType() == RECURSIVE) {

			if (initial == null)
				throw new InputException(Messages.getString(BUNDLE_NAME,
						"Sequence.4")); //$NON-NLS-1$

			if (initial.length < sequencedepth) {
				if (sequencedepth > 1)
					throw new InputException(Messages.getString(BUNDLE_NAME,
							"Sequence.5") //$NON-NLS-1$
							+ String.valueOf(sequencedepth)
							+ Messages.getString(BUNDLE_NAME, "Sequence.6")); //$NON-NLS-1$
			}

			if (initial.length > sequencedepth) {
				if (sequencedepth > 1)
					throw new InputException(Messages.getString(BUNDLE_NAME,
							"Sequence.7") //$NON-NLS-1$
							+ String.valueOf(sequencedepth)
							+ Messages.getString(BUNDLE_NAME, "Sequence.8")); //$NON-NLS-1$
				else
					throw new InputException(Messages.getString(BUNDLE_NAME,
							"Sequence.9")); //$NON-NLS-1$

			}
		} else {
			if (!(initial == null))
				throw new InputException(Messages.getString(BUNDLE_NAME,
						"Sequence.10")); //$NON-NLS-1$
		}
	}

	public String getFormula() {
		return formula.toString();
	}

	/**
	 * berechnet die Folge. Eine Exception wird bei ?berlauf/Division durch Null
	 * geworfen
	 */
	public double[] compute(double[] initials, int startindex, int endindex)
			throws Exception {

		this.initial = initials;
		this.startindex = startindex;
		this.endindex = endindex;

		check();

		double[] sequence = null;

		//rekursive Folgen
		if (sequencedepth > 0) { // Startwerte einsetzen
			if (endindex > this.maxlengthsequence) {
				throw new Exception(Messages.getString(BUNDLE_NAME,
						"Sequence.11") //$NON-NLS-1$
						+ this.maxlengthsequence);
			}

			if (endindex < sequencedepth) {
				sequence = new double[endindex + 1];
				for (int i = 0; i < endindex + 1; i++) {
					sequence[i] = initial[i];
				}
				return sequence;
			}

			int pos = sequencedepth;
			sequence = new double[endindex + 1];
			for (int i = 0; i < pos; i++) {

				sequence[i] = initial[i];
			}

			while (pos < endindex + 1) {
				//Wert f?r explizite Variable setzten
				((Variable) variables.get("n")).setValue(pos); //$NON-NLS-1$

				for (int i = 0; i < rekindices.size(); i++) {
					String varName = "a(n-" //$NON-NLS-1$
							+ String.valueOf(rekindices.elementAt(i)) + ")"; //$NON-NLS-1$

					Variable v = (Variable) variables.get(varName);
					Integer k = (Integer) rekindices.elementAt(i);
					double a = sequence[pos - k.intValue()];
					v.setValue(a);
				}
				sequence[pos] = formula.value();
				pos++;
			}
		} else { //explizite Folgen

			if (endindex - startindex > this.maxlengthsequence) {
				throw new Exception(Messages.getString(BUNDLE_NAME,
						"Sequence.15") //$NON-NLS-1$
						+ this.maxlengthsequence);
			}
			sequence = new double[endindex - startindex + 1];
			int pos = 0;

			while (pos < endindex - startindex + 1) {
				//neuestes Folgenglied berechnen
				//Wert f?r explizite Variable setzten
				((Variable) variables.get("n")).setValue(pos + startindex); //$NON-NLS-1$
				sequence[pos] = formula.value();
				pos++;
			}
		}

		//auf Infinity, NaN pr?fen
		for (int i = 0; i < sequence.length; i++) {
			if (Double.isInfinite(sequence[i])) {

				throw new ComputationException(Messages.getString(BUNDLE_NAME,
						"Sequence.17") //$NON-NLS-1$
						+ Messages.getString(BUNDLE_NAME, "Sequence.18"));} //$NON-NLS-1$

			if (Double.isNaN(sequence[i])) {
				throw new ComputationException(Messages.getString(BUNDLE_NAME,
						"Sequence.19"));} //$NON-NLS-1$

		}

		//falls wir rekursive berechnet haben m?ssen wir noch den Anfang
		// wegmachen
		if (sequencedepth > 0) {
			double[] vals = new double[endindex - startindex + 1];
			for (int i = startindex; i <= endindex; i++) {
				vals[i - startindex] = sequence[i];

			}
			return vals;

		} else {
			return sequence;
		}

	}

	/**
	 * Gibt den Typ der Folge (rekursiv oder explizit) zur?ck
	 * 
	 * @return Typ der Folge als Klassenkonstante RECURSIVE oder EXPLICIT
	 */
	public int getType() {
		return sequencedepth > 0 ? RECURSIVE : EXPLICIT;
	}

}