package uibk.mtk.math;

import uibk.mtk.math.functions.Function1D;
import uibk.mtk.util.IntervalIterator;

public class SingularityChecker {

	static int SINGCHECK_DISCRETSIZE = 30;

	/**
	 * Sicherheitsfaktor. Dieser Faktor wird mit (zmax-zmin) multipliert und das
	 * Ergebnis muss von der aktuellen Differenz ?bertroffen werden damit wir
	 * auf Singularit?t erkennen
	 */
	static final double FACMIN = 0.001;

	/**
	 * Dieser Faktor ist f?r das zweite Testkriterium. Falls die aktuelle
	 * Differenz um diesen Faktor gr??er ist als die durchschnittliche bis zu
	 * diesem Zeitpunkt berechnete durchschnittliche Differenz -> Singularit?t
	 *  
	 */
	static final double FACDIFF = 10;

	/**
	 * Schwellwert ab dem sicher eine Singularit?t vorliegt (siehe
	 * THRESHOLDFACTOR)
	 */
	double thresh_sure;

	/**
	 * Dieser Faktor wird mit (zmax-zmin) multipliziert und das Erbegnis ist ein
	 * "sicheres" Kriterium (falls der Faktor nicht zu klein ist) ob zw. zwei
	 * Punkten eine Singularit?t vorhanden ist. (mit diesem Test findet man
	 * leich Sachen wie 1/x)
	 */
	static final double THRESHOLDFACTOR = 0.3;

	/** Schwellwert der mindestens ?bertroffen werden muss (f?r 2.Testkriterium) */
	double thresh_min;

	int discretsize = SINGCHECK_DISCRETSIZE;

	public SingularityChecker(double min, double max) {

		thresh_sure = (max - min) * THRESHOLDFACTOR;
		thresh_min = Math.max(FACMIN * (max - min), 0.0001);

	}

	public SingularityChecker(double min, double max, int size) {

		this(min, max);
		discretsize = size;

	}

	/**
	 * Testen ob auf der Strecke zw. den Punkten p1 und p2 eine Singularit?t
	 * auftritt (Richtung spielt eine Rolle !)
	 *  
	 */
	public boolean hasSingularity(Function1D f, double xmin, double xmax) {

		IntervalIterator it1 = new IntervalIterator(xmin, xmax, discretsize);
		IntervalIterator it2 = new IntervalIterator(xmax, xmin, discretsize);
		if (hassing(it1, f) || hassing(it2, f))
			return true;
		else
			return false;
	}

	private boolean hassing(IntervalIterator it, Function1D f) {
		double last = 0;

		//Summe der Differenzen
		double sum = 0;
		int c = 0;

		while (it.hasNext()) {

			double t = it.nextdouble();
			double z = f.getValue(t);

			if (Double.isNaN(z) || Double.isInfinite(z))
				return true;

			if (it.getPos() > 1) {

				double diff = Math.abs(z - last);

				// 1.Kriterum: ist die Differenz sehr gro? -> Singularit?t
				if (diff > thresh_sure)
					return true;

				if (c > 0 && sum > thresh_min && diff > (sum / c) * FACDIFF) {

					return true;

				}
				sum += diff;
				c++;

			}
			last = z;
		}

		return false;
	}

}