package uibk.mtk.math.functions;

import uibk.mtk.lang.Messages;
import uibk.mtk.lang.Settings;

/**
 * Klasse f?r eine reelle Funktion von R^n->R^m
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public abstract class Function {

	/** Name des Resourcebundles */
	private static final String BUNDLE_NAME = "uibk.mtk.math.functions.messages";//$NON-NLS-1$

	/** Anzahl der Variablen */
	protected int n;

	/** Anzahl der Gleichungen */
	protected int m;

	/**
	 * Sicherheitsfaktor>=1 f?r das Testen auf Diffenzierbarkeit, je kleiner der
	 * Faktor desto eher werden differnzierbare Stellen als nicht
	 * differenzierbar "gewertet"
	 */
	static final double SAFETYFACTOR = 1000.0;

	/** Register */
	private double[] temp;

	/** Register f?r Vorw?rtsdifferenz */
	private double[] forwarddiff;

	/** Register f?r R?ckw?rtsdifferenz */
	private double[] backwarddiff;

	/**
	 * setzt fest ob bei der Berechnung der Ableitungen auf Differenzierbarkeit
	 * getestet werden soll. Der Standardwert ist <CODE>true</CODE>. Falls
	 * dies aktiviert ist wird falls die Fkt. als nicht differenzierbar gewertet
	 * wird eine <CODE>DerivationException</CODE> geworfen
	 */
	protected boolean checkdiff = true;

	/**
	 * Erzeugt eine Funktio von R^n->R^m
	 * 
	 * @param n
	 *            Anzahl der Variablen >0
	 * @param m
	 *            Anzahl der Komponenten >0
	 */
	public Function(int n, int m) {
		this.m = m;
		this.n = n;
		temp = new double[n];

		forwarddiff = new double[m];
		backwarddiff = new double[m];

	}

	/** Gibt die Anzahl der Variablen der Funktion zur?ck */
	public int getNumberVars() {
		return n;
	}

	/** Gibt die Anzahl der Gleichungen der Funktion zur?ck */
	public int getNumberFunctions() {
		return m;
	}

	/**
	 * Berechnung des Wertes der Funktion an der Stelle x
	 * 
	 * @return Wert der Funktion
	 */
	public abstract double[] getValue(double[] x);

	/**
	 * gibt die Differenz h zum Berechnen der 1.Ableitung f?r die symmetrische
	 * Differenzenformel zur?ck
	 * 
	 * @param x
	 *            Stelle an der differenziert wird
	 *  
	 */
	private double gethSymm1(double x) {
		//h f?r die symmetrischen Diff. berechnen
		double hsym = Math.pow(Settings.getMachinePrecision(), 1.0 / 3);

		double s = Math.max(1, Math.abs(x));
		return hsym * s;
	}

	/**
	 * gibt die Differenz h zum Berechnen der 1.Ableitung f?r die einseitigen
	 * Differenzenformelen zur?ck
	 * 
	 * @param x
	 *            Stelle an der differenziert wird
	 *  
	 */
	private double getHForBack1(double x) {
		//h f?r die symmetrischen Diff. berechnen
		double hsym = Math.pow(Settings.getMachinePrecision(), 1.0 / 2);

		double s = Math.max(1, Math.abs(x));
		return hsym * s;
	}

	/**
	 * gibt die Differenz h zum Berechnen der 2.Ableitung f?r die symmetrische
	 * Differenzenformel zur?ck
	 * 
	 * @param x
	 *            Stelle an der differenziert wird
	 *  
	 */
	private double gethSymm2(double x) {
		//h f?r die symmetrischen Diff. berechnen
		double hsym = Math.pow(Settings.getMachinePrecision(), 1.0 / 4);

		double s = Math.max(1, Math.abs(x));
		return hsym * s;
	}

	/**
	 * gibt die Differenz h zum Berechnen der 2.Ableitung f?r die einseitigen
	 * Differenzenformelen zur?ck
	 * 
	 * @param x
	 *            Stelle an der differenziert wird
	 *  
	 */
	private double getHForBack2(double x) {
		//h f?r die symmetrischen Diff. berechnen
		double hsym = Math.pow(Settings.getMachinePrecision(), 1.0 / 3);

		double s = Math.max(1, Math.abs(x));
		return hsym * s;
	}

	/**
	 * Berechnung der Jacobimatrix an der Stelle <CODE>x</CODE>
	 * 
	 * @param x
	 *            Wert an der differenziert werden soll
	 * @return die Jacobimatrix
	 * @throws DerivationException
	 *             falls auf Differenzierbarkeit testen aktiv ist und die
	 *             Funktion als nicht differenzierbar "erkannt" wurde
	 */
	public double[][] getJacobian(double[] x) throws DerivationException {

		if (x.length != n)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"Function.0")); //$NON-NLS-1$

		//Feld f?r Jacobimatrix
		double[][] jacob = new double[m][n];

		//Eintr?ge von x nach temp kopieren
		System.arraycopy(x, 0, temp, 0, n);

		//h f?r die Vorw?rts/R?ckw?rtdiff
		// double h = Math.sqrt(Settings.getMachinePrecision());

		for (int j = 0; j < n; j++) {

			//falls aktiviert auf Diffbarkeit ?berpr?fen
			if (checkdiff) {
				check_derivate1(x[j]);
			}
			//Berechnung der Funktionswerte vorw?rts und r?ckw?rts
			double t = x[j];

			double h = gethSymm1(t);

			temp[j] = t + h;
			double[] forward = getValue(temp);
			temp[j] = t - h;
			double[] backward = getValue(temp);
			temp[j] = t;

			//Berechnung der zentralen Differenzen
			for (int k = 0; k < m; k++) {
				jacob[k][j] = (forward[k] - backward[k]) / (2 * h);
				if (checkdiff
						&& (Double.isInfinite(jacob[k][j]) || Double
								.isNaN(jacob[k][j])))
					throw new DerivationException(Messages.getString(
							BUNDLE_NAME, "Function.1")); //$NON-NLS-1$
			}

		}
		return jacob;

	}

	/**
	 * ?berpr?fen auf Differenzierkeit der 2.Ableitung im Punkt <CODE>x</CODE>
	 */
	private void check_derivate2(double x) throws DerivationException {

		double h = getHForBack2(x);

		//verwendete Vorw?rtsdifferenz
		// ((2*y(x0)-5*y(x0+h)+4*y(x0+2*h)-y(x0+3*h))/h/h,h);
		//verwendete R?ckw?rtssdifferenz
		// ((2*y(x0)-5*y(x0-h)+4*y(x0-2*h)-y(x0-3*h))/h/h,h);

		double[] center = getValue(new double[] { x });
		double[] forward = getValue(new double[] { x + h });
		double[] backward = getValue(new double[] { x - h });
		double[] forward2 = getValue(new double[] { x + 2 * h });
		double[] backward2 = getValue(new double[] { x - 2 * h });
		double[] forward3 = getValue(new double[] { x + 3 * h });
		double[] backward3 = getValue(new double[] { x - 3 * h });

		double forwardDiff, backwardDiff;

		for (int k = 0; k < m; k++) {
			forwardDiff = (2 * center[k] - 5 * forward[k] + 4 * forward2[k] - forward3[k])
					/ (h * h);
			backwardDiff = (2 * center[k] - 5 * backward[k] + 4 * backward2[k] - backward3[k])
					/ (h * h);

			double scale = Math.max(1, Math.max(Math.abs(forward[k]), Math
					.abs(backward[k])));
			double estimated_acurracy = Math.pow(h, 2.0 / 3.0) * scale;

			if (Double.isInfinite(forwardDiff)
					|| Double.isNaN(forwardDiff)
					|| Double.isInfinite(backwardDiff)
					|| Double.isNaN(backwardDiff)
					|| (Math.abs(forwardDiff - backwardDiff)) > SAFETYFACTOR
							* estimated_acurracy) {
				throw new DerivationException(Messages.getString(BUNDLE_NAME,
						"Function.2"));} //$NON-NLS-1$
		}
	}

	/**
	 * ?berpr?fen auf Differenzierkeit der 1.Ableitung im Punkt <CODE>x</CODE>
	 */
	void check_derivate1(double t) throws DerivationException {

		for (int j = 0; j < n; j++) {

			//Berechnung der Funktionswerte vorw?rts und r?ckw?rts

			double h = getHForBack1(t);
			temp[j] = t + h;
			double[] forward = getValue(temp);
			temp[j] = t - h;
			double[] backward = getValue(temp);
			temp[j] = t;
			double[] center = getValue(temp);

			for (int k = 0; k < m; k++) { //Berechnung der Vorw?rts
				// undR?ckw?rtsdifferenzen

				forwarddiff[k] = (forward[k] - center[k]) / h;
				backwarddiff[k] = (center[k] - backward[k]) / h;

				double scale = Math.max(1, Math.max(Math.abs(forward[k]), Math
						.abs(backward[k])));

				double estimated_acurracy = Math.sqrt(h) * scale;
				//* Math.max(Math.sqrt(Math.abs(center[k])), 1);

				if (Double.isInfinite(backward[k])
						|| Double.isNaN(backward[k])
						|| Double.isInfinite(forward[k])
						|| Double.isNaN(forward[k])
						|| (Math.abs(forwarddiff[k] - backwarddiff[k])) > SAFETYFACTOR
								* estimated_acurracy)
					throw new DerivationException(Messages.getString(
							BUNDLE_NAME, "Function.3")); //$NON-NLS-1$
			}

		}
	}

	/**
	 * Berechnet die zweite Ableitung an der Stelle <CODE>y</CODE>. Dies ist
	 * bis jetzt nur f?r Funktionen von R->R^m implementiert
	 * 
	 * @param y
	 *            Wert an dem differeniert werden soll
	 * @throws DerivationException
	 *             falls auf Differenzierbarkeit testen aktiv ist und die
	 *             Funktion als nicht 2.differenzierbar "erkannt" wird
	 */

	protected double[] getDerivation2(double[] y) throws DerivationException {

		if (n != 1)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"Function.4")); //$NON-NLS-1$

		if (y.length != 1)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"Function.5")); //$NON-NLS-1$

		double x = y[0];

		if (checkdiff) {
			check_derivate2(x);
		}

		double h = this.gethSymm2(x);

		double[] ret = new double[m];

		//verwendete Differenzenformel f(x+h)-2*f(x)+f(x-h))/h^2
		double[] center = getValue(y);
		//Vorw?rts
		double[] forward = getValue(new double[] { x + h });
		//R?ckw?rts
		double[] backward = getValue(new double[] { x - h });

		for (int k = 0; k < m; k++) {
			ret[k] = (forward[k] - 2 * center[k] + backward[k]) / (h * h);
			if (checkdiff)
				if (Double.isInfinite(ret[k]) || Double.isInfinite(ret[k]))
					throw new DerivationException(Messages.getString(
							BUNDLE_NAME, "Function.6")); //$NON-NLS-1$
		}

		return ret;

	}

	/** Gibt die gesch?tzte Genauigkeit f?r den Fehler der 1.Ableitung zur?ck */
	protected double getEstimatedErrorFirstDeriv(double x) {
		double h = this.gethSymm1(x);
		return Math.pow(h, 2.0 / 3.0);
	}

	/** Gibt die gesch?tzte Genauigkeit f?r den Fehler der 1.Ableitung zur?ck */
	protected double getEstimatedErrorSecondDeriv(double x) {
		double h = this.gethSymm2(x);
		return Math.pow(h, 2.0 / 3.0);
	}

	/**
	 * Berechnung der Richtungableitung in <CODE>x</CODE> in Richtung <CODE>
	 * dir</CODE>
	 * 
	 * @param x
	 *            Punkt
	 * @param dir
	 *            Richtung
	 * @return
	 * @throws DerivationException
	 *             falls auf Differenzierbarkeit testen aktiv ist und die
	 *             Funktion als nicht differenzierbar "erkannt" wird
	 */
	public double[] directionalDerivate(double[] x, double[] dir)
			throws DerivationException {
		if (x.length != n)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"Function.7")); //$NON-NLS-1$

		double[] v = (double[]) dir.clone();

		double sum = 0;
		for (int i = 0; i < n; i++) {
			sum = sum + v[i] * v[i];
		}
		sum = Math.sqrt(sum);
		for (int i = 0; i < n; i++) {
			v[i] /= sum;
		}

		if (checkdiff)
			check_directional(x, v);

		double h = this.gethSymm1(x[0]);

		//Vorw?rts, R?ckw?rts und zentrale Differenz berechnen
		double[] ftemp = new double[n];
		double[] btemp = new double[n];
		for (int i = 0; i < n; i++) {

			ftemp[i] = x[i] + h * v[i];
			btemp[i] = x[i] - h * v[i];
		}

		double[] forward = getValue(ftemp);

		double[] backward = getValue(btemp);

		double[] deriv = new double[m];

		//Berechnung der zentralen Differenzen
		for (int k = 0; k < m; k++) {

			deriv[k] = (forward[k] - backward[k]) / (2 * h);
			if (checkdiff
					&& (Double.isInfinite(deriv[k]) || Double.isNaN(deriv[k])))
				throw new DerivationException(Messages.getString(BUNDLE_NAME,
						"Function.8")); //$NON-NLS-1$
		}

		return deriv;

	}

	/** ?berpr?fen der Richtungsableitunf auf Differenzierbarkeit */
	private void check_directional(double[] x, double[] v)
			throws DerivationException {

		double h = this.getHForBack1(x[0]);

		//Vorw?rts, R?ckw?rts und zentrale Differenz berechnen
		double[] ftemp = new double[n];
		double[] btemp = new double[n];
		for (int i = 0; i < n; i++) {

			ftemp[i] = x[i] + h * v[i];
			btemp[i] = x[i] - h * v[i];
		}

		double[] forward = getValue(ftemp);
		double[] center = getValue(x);
		double[] backward = getValue(btemp);

		for (int k = 0; k < m; k++) { //Berechnung der Vorw?rts
			// undR?ckw?rtsdifferenzen

			forwarddiff[k] = (forward[k] - center[k]) / h;
			backwarddiff[k] = (center[k] - backward[k]) / h;

			double estimated_acuuracy = Math.sqrt(h);

			if (Double.isInfinite(backward[k])
					|| Double.isNaN(backward[k])
					|| Double.isInfinite(forward[k])
					|| Double.isNaN(forward[k])
					|| (Math.abs(forwarddiff[k] - backwarddiff[k])) > SAFETYFACTOR
							* estimated_acuuracy)
				throw new DerivationException(Messages.getString(BUNDLE_NAME,
						"Function.9")); //$NON-NLS-1$
		}

	}

}