package uibk.mtk.math.parsing;

import java.util.StringTokenizer;
import java.util.Vector;

import uibk.mtk.geom.geom2d.CoordinateRect2D;
import uibk.mtk.geom.geom2d.Punkt2D;
import uibk.mtk.lang.InputException;
import uibk.mtk.lang.Messages;
import uibk.mtk.math.Interval;
import uibk.mtk.parser.realparser.Expr;
import uibk.mtk.parser.realparser.Parser;

/**
 * Utilityklasse die einige Funktionen zum Parsen von math. Ausdr?cken wie
 * Intervallen, Anfangsbedingungen usw. bereitstellt
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2003
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public final class Misc {

	private Misc() {
	}

	/** Name des Resourcebundles */
	private static final String BUNDLE_NAME = "uibk.mtk.math.parsing.messages";//$NON-NLS-1$

	/**
	 * Das zum Parsen von vorkommenden mathematische Ausdr?cken verwendete
	 * Parser-Objekt
	 */
	private static Parser parser = new Parser();

	/**
	 * Funktion die einen konstanten mathematischen Ausdruck wie z.B. sin(2)
	 * oder 2*pi auswertet.
	 * 
	 * @param expr
	 *            ein konstanter mathematischer Ausdruck
	 * @return der Wert des Ausdrucks
	 * @throws InputException
	 *             die vom Parser geworfene InputException
	 */
	public static synchronized double parseConstantExpr(String expr)
			throws InputException {
		Expr e;

		e = parser.parseString(expr);

		return e.value();
	}

	/**
	 * Wertet ein durch einen String der Form "[a,b]" definiertes Intervall aus.
	 * 
	 * @param strInterval
	 *            ein String der Form "[a,b]"
	 * @param name
	 *            Name des Intervalls (wird bei den Fehlermeldungen verwendet)
	 * @param trueinterval
	 *            gibt an ob es ein echtes Intervall sein muss (falls ja und
	 *            Intervall ist nicht echt -> InputException)
	 * 
	 * @throws InputException
	 */
	public static synchronized Interval parseIntervall(String strInterval,
			String name, boolean trueinterval) throws InputException {

		if (name == null)
			name = ""; //$NON-NLS-1$

		if (strInterval == null || strInterval.equals("")) //$NON-NLS-1$
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.2")); //$NON-NLS-1$

		String str = strInterval.trim();

		//?berpr?fen ob Intervall mit "[" beginnt und mit "]" endet
		if (!str.startsWith("[")) //$NON-NLS-1$
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.4")); //$NON-NLS-1$
		if (!str.endsWith("]")) //$NON-NLS-1$
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.6")); //$NON-NLS-1$

		//Positon von "," bestimmen
		int pos = str.indexOf(","); //$NON-NLS-1$
		if (pos < 0)
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.8")); //$NON-NLS-1$

		double start, end;

		//untere Grenze parsen
		try {
			start = parseConstantExpr(str.substring(1, pos));
		} catch (InputException ex) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.9")); //$NON-NLS-1$
		}

		//oebere Grenze parsen
		try {
			end = parseConstantExpr(str.substring(pos + 1,
					strInterval.length() - 1));
		} catch (InputException ex) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.10")); //$NON-NLS-1$
		}

		//falls gew?nscht auf echtes Intervall pr?fen
		if (trueinterval) {
			if (end <= start)
				throw new InputException(name
						+ Messages.getString(BUNDLE_NAME, "Misc.11")); //$NON-NLS-1$
		} else {
			//ansonsten pr?fen ob a<=b
			if (end < start)
				throw new InputException(name
						+ Messages.getString(BUNDLE_NAME, "Misc.12")); //$NON-NLS-1$
		}

		//Intervall zur?ckgeben
		Interval interval = new Interval(start, end);
		return interval;
	}

	/**
	 * Funktion zum Parsen von Anfangswerten, die in der Form 2,4,Pi,sqrt(2)
	 * ?bergeben werden
	 * 
	 * @param strinitials
	 *            konstante mathematische Ausdr?cke, durch Beistriche getrennt
	 * 
	 * @param name
	 *            Name der Werte (f?r Fehlermeldungen)
	 * @return ein Array mit den Werten der Ausdr?cke
	 * 
	 * @throws InputException
	 */
	public static synchronized double[] parseInitalValues(String strinitials,
			String name) throws InputException {

		if (strinitials == null || strinitials.equals("")) //$NON-NLS-1$
			throw new InputException(
					Messages.getString(BUNDLE_NAME, "Misc.14") + name + Messages.getString(BUNDLE_NAME, "Misc.15")); //$NON-NLS-1$ //$NON-NLS-2$

		//trimmen
		String str = strinitials.trim();

		//?berpr?fen ob die Startwerte die richtige Form haben
		if (str.charAt(0) == ',' || str.charAt(strinitials.length() - 1) == ','
				|| str.indexOf(",,") >= 0) //$NON-NLS-1$
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.17")); //$NON-NLS-1$

		//String zerteilen
		String[] strvalues = split(str, ","); //$NON-NLS-1$

		if (strvalues == null || strvalues.length == 0)
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.19")); //$NON-NLS-1$

		//Auswerten der Werte
		double[] values = new double[strvalues.length];

		for (int i = 0; i < values.length; i++) {
			try {
				values[i] = parseConstantExpr(strvalues[i]);
			} catch (InputException ex) {
				throw new InputException(Messages.getString(BUNDLE_NAME,
						"Misc.20") //$NON-NLS-1$
						+ name + Messages.getString(BUNDLE_NAME, "Misc.21")); //$NON-NLS-1$
			}
		}

		//Werte zur?ckgeben
		return values;

	}

	/**
	 * Gleich wie die Funktion <CODE>parseInitalValues</CODE>, nur das die
	 * Werte in Klammern stehen m?ssen.
	 * 
	 * @param strinitials
	 *            z.b.die in der Form (2,4,Pi,sqrt(2))
	 * 
	 * @param name
	 *            Name der Werte (f?r Fehlermeldungen)
	 * 
	 * @return
	 * @throws InputException
	 */
	public static synchronized double[] parseBracketInitalValues(
			String strinitials, String name) throws InputException {

		String str = strinitials.trim();

		if (str.charAt(0) == '(' && str.charAt(str.length() - 1) == ')') {
			str = str.substring(1, str.length() - 1);
		} else {
			try {
				double[] ret = new double[1];
				ret[0] = parseConstantExpr(str);
				return ret;
			} catch (InputException ex) {
				throw new InputException(name
						+ Messages.getString(BUNDLE_NAME, "Misc.22")); //$NON-NLS-1$
			}
		}
		return parseInitalValues(str, name);
	}

	/**
	 * Zerteilt einen String in Teile die durch eine andere Zeichenfolge
	 * getrennt werden
	 * 
	 * @param aString
	 *            zu zerteilender String z.B. "1,2,3"
	 * @param delimiter
	 *            zum Trennen zu verwendender String z.B. ","
	 */
	public static synchronized String[] split(String aString, String delimiter) {
		StringTokenizer st = new StringTokenizer(aString, delimiter);
		String[] result = new String[st.countTokens()];
		int i = 0;
		while (st.hasMoreTokens()) {
			result[i++] = st.nextToken();
		}
		return result;
	}

	/**
	 * Funktion zum parsen einer Genauigkeit
	 * 
	 * @param streps
	 *            ein String der eine Zahl enth?lt
	 * @return double die durch den String definierte Gleitkommazahl
	 * @throws InputException
	 *             falls keine Zahl,String leer, oder Zahl zu klein
	 */
	public static synchronized double parseEps(String streps, double min)
			throws InputException {

		return parseDoubleValue(streps, Messages.getString(BUNDLE_NAME,
				"Misc.23"), //$NON-NLS-1$
				new Double(min), null, Boolean.valueOf(true));
	}

	/**
	 * Funktion zum Parsen einer reellen Zahl.
	 * 
	 * @param strvalue
	 *            ein String der eine Zahl enth?lt
	 * @param name
	 *            Name des Wertes f?r Fehlermeldungen
	 * @param min
	 *            falls min!=null wird ?berpr?ft ob der Wert gr??er gleich min
	 * @param max
	 *            falls max!=null wird ?berpr?ft ob der Wert kleiner gleich max
	 * @param positive
	 *            falls positive!=null wird,falls true ?berpr?ft ob der Wert
	 *            gr??er 0 und falls false ob der Wert kleiner 0 ist. (geht zwar
	 *            auch mit min und max aber hier kommt entrsprechende Meldung
	 *            ("muss positiv/negativ sein")
	 * 
	 * @return die durch den String definierte Gleitkommazahl
	 * @throws InputException
	 *             InputException falls keine Zahl,String leer, oder Zahl zu
	 *             klein oder nicht im Berecih
	 */
	public static synchronized double parseDoubleValue(String strvalue,
			String name, Double min, Double max, Boolean positive)
			throws InputException {
		double val = 0;
		if (strvalue == null || strvalue.equals("")) //$NON-NLS-1$
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.25")); //$NON-NLS-1$
		try {
			val = parseConstantExpr(strvalue);
		} catch (InputException ex) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.26")); //$NON-NLS-1$
		}
		if (positive != null && positive.booleanValue() && val <= 0) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.27")); //$NON-NLS-1$
		}
		if (positive != null && !positive.booleanValue() && val >= 0) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.28")); //$NON-NLS-1$
		}
		if (max != null && val > max.doubleValue()) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.29") //$NON-NLS-1$
					+ max + Messages.getString(BUNDLE_NAME, "Misc.30")); //$NON-NLS-1$
		}
		if (min != null && val < min.doubleValue()) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.31") //$NON-NLS-1$
					+ min + Messages.getString(BUNDLE_NAME, "Misc.32")); //$NON-NLS-1$
		}
		return val;
	}

	/**
	 * Funktion zum Parsen einer ganzen Zahl.
	 * 
	 * @param strvalue
	 *            ein String der eine Zahl enth?lt
	 * @param name
	 *            Name des Wertes f?r Fehlermeldungen
	 * @param min
	 *            falls min!=null wird ?berpr?ft ob der Wert gr??er gleich min
	 * @param max
	 *            falls max!=null wird ?berpr?ft ob der Wert kleiner gleich max
	 * 
	 * @return die durch den String definierte Gleitkommazahl
	 * @throws InputException
	 *             InputException falls keine Zahl,String leer, oder Zahl zu
	 *             klein oder nicht im Berecih
	 */
	public static synchronized int parseIntegerValue(String strvalue,
			String name, Integer min, Integer max) throws InputException {

		int val = 0;

		boolean natural = false;
		if (min != null && min.intValue() >= 0)
			natural = true;

		if (strvalue == null || strvalue.equals("")) //$NON-NLS-1$
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.34")); //$NON-NLS-1$
		try {
			val = Integer.parseInt(strvalue);
		} catch (Exception ex) {

			String type = natural ? Messages.getString(BUNDLE_NAME, "Misc.35") : Messages.getString(BUNDLE_NAME, "Misc.36"); //$NON-NLS-1$ //$NON-NLS-2$
			throw new InputException(
					name
							+ Messages.getString(BUNDLE_NAME, "Misc.37") + type + Messages.getString(BUNDLE_NAME, "Misc.38")); //$NON-NLS-1$ //$NON-NLS-2$
		}
		if (natural && val < 0) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.39")); //$NON-NLS-1$
		}
		if (max != null && max.intValue() < 0 && val >= 0) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.40")); //$NON-NLS-1$
		}
		if (max != null && val > max.doubleValue()) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.41") //$NON-NLS-1$
					+ max + Messages.getString(BUNDLE_NAME, "Misc.42")); //$NON-NLS-1$
		}
		if (min != null && val < min.doubleValue()) {
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.43") //$NON-NLS-1$
					+ min + Messages.getString(BUNDLE_NAME, "Misc.44")); //$NON-NLS-1$
		}
		return val;
	}

	/**
	 * Funktion zum Parsen eines durch vier Zeichenketten (die Grenzen des
	 * Rechtecks) gegebenen Rechtecks
	 */
	public static synchronized CoordinateRect2D parseRectArea(String strxmin,
			String strxmax, String strymin, String strymax)
			throws InputException {

		double xmin = 0, xmax = 0, ymin = 0, ymax = 0;

		//?berpr?fen ob alles nichtleer
		if (strxmin == null || strxmin.equals("")) //$NON-NLS-1$
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.46")); //$NON-NLS-1$
		if (strxmax == null || strxmax.equals("")) //$NON-NLS-1$
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.48")); //$NON-NLS-1$
		if (strymin == null || strymin.equals("")) //$NON-NLS-1$
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.50")); //$NON-NLS-1$
		if (strymax == null || strymax.equals("")) //$NON-NLS-1$
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.52")); //$NON-NLS-1$

		//xmin parsen
		try {
			xmin = parseConstantExpr(strxmin);
		} catch (InputException e) {
			throw new InputException(
					Messages.getString(BUNDLE_NAME, "Misc.53") + e.getMessage()); //$NON-NLS-1$
		}

		//xmax parsen
		try {
			xmax = parseConstantExpr(strxmax);
		} catch (InputException e) {
			throw new InputException(
					Messages.getString(BUNDLE_NAME, "Misc.54") + e.getMessage()); //$NON-NLS-1$
		}

		//ymin parsen
		try {
			ymin = parseConstantExpr(strymin);
		} catch (InputException e) {
			throw new InputException(
					Messages.getString(BUNDLE_NAME, "Misc.55") + e.getMessage()); //$NON-NLS-1$
		}

		//ymax parsen
		try {
			ymax = parseConstantExpr(strymax);
		} catch (InputException e) {
			throw new InputException(
					Messages.getString(BUNDLE_NAME, "Misc.56") + e.getMessage()); //$NON-NLS-1$
		}

		if (xmax <= xmin)
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.57")); //$NON-NLS-1$
		if (ymax <= ymin)
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.58")); //$NON-NLS-1$

		return new CoordinateRect2D(xmin, xmax, ymin, ymax);
	}

	/**
	 * Parsen eines 2D-Punktes.
	 * 
	 * @param strpoint
	 *            ein String der Form "(x,y)"
	 * @return ein Punkt2D mit den durch <CODE>strpoint</CODE> def.
	 *         Koordinaten
	 * @throws InputException
	 */
	public static Punkt2D parsePunkt2D(String strpoint, String name)
			throws InputException {

		if (strpoint == null || strpoint.equals("")) { //$NON-NLS-1$
			throw new InputException(name
					+ Messages.getString(BUNDLE_NAME, "Misc.60")); //$NON-NLS-1$

		}

		try {
			double[] coords = parseBracketInitalValues(strpoint, ""); //$NON-NLS-1$
			return new Punkt2D(coords[0], coords[1]);
		} catch (InputException e) {
			throw new InputException(Messages.getString(BUNDLE_NAME, "Misc.62") //$NON-NLS-1$
					+ name + Messages.getString(BUNDLE_NAME, "Misc.63")); //$NON-NLS-1$
		}
	}

	/***************************************************************************
	 * Liest aus einer Zeichenkette der Form x1,cos(x2)*sin(x11),x5 alle Indice
	 * der Variablen aus (hier "x").
	 * 
	 * @param del
	 *            Zeichen das die einzelnen Funktionen trennt (z.B. ",")
	 * @param varname
	 *            die Variable (z.B. "x")
	 * @param str
	 *            String mit ausdr?cken in x1,x2,....
	 * 
	 * @return Feld mit den Indices oder leeres Feld falls str leer ist oder
	 *         keine Indices gefunden worden sind (d.h. falls auch die Variable
	 *         selbst nicht vorkommt ansonsten wird eine InputException
	 *         geworfen)
	 * @throws InputException
	 */
	public static int[] parseVarsIndices(String del, String varname, String str) {

		if (str == null)
			return new int[] {};

		//Splitten der Eingabe
		String[] v = split(str.trim(), del);

		//leeres Feld zur?ckgeben
		if (v == null || v.length == 0)
			return new int[] {};

		//Die Indice der Variablen
		Vector indices = new Vector();

		for (int i = 0; i < v.length; i++) {

			int pos = -1;

			do {
				pos = v[i].indexOf(varname, ++pos);
				if (pos == -1)
					break;

				int index = scanNumber(pos + varname.length(), v[i]);

				//?berpr?fen ob schon hinzugef?gt
				if (index != -1) {
					boolean add = true;
					for (int k = 0; k < indices.size(); k++) {
						if (((Integer) indices.elementAt(k)).intValue() == index)
							add = false;
					}

					if (add)
						indices.add(new Integer(index));
				}

			} while (pos >= 0);

		}

		//aus Vector mit Integer ein int[] Feld machen
		int[] ind = new int[indices.size()];
		for (int i = 0; i < indices.size(); i++) {
			ind[i] = ((Integer) indices.elementAt(i)).intValue();
		}

		return ind;
	}

	/**
	 * Ausgehend von der Position <CODE>i</CODE> wird im String nach einer
	 * m?glichst langen Zahl gesucht und diese Zahl zur?ckgegeben.
	 * 
	 * @param i
	 *            Startposition im String
	 * @param s
	 *            ein String
	 * 
	 * @return -1 falls keine Zahl gefunden wird ansonsten die Zahl
	 */
	private static int scanNumber(int i, String s) {

		if (i >= s.length())
			return -1;
		int from = i;

		for (; i < s.length(); ++i)
			if (!Character.isDigit(s.charAt(i)))
				break;

		if (from == i || i > s.length())
			return -1;

		String text = s.substring(from, i);
		int nval;

		nval = Integer.valueOf(text).intValue();
		return nval;

	}

}