package uibk.mtk.text;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;

/**
 * Klasse mit Funktionen zur formattierten Darstellung von Dezimalzahlen
 * 
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2004
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public class Formatter extends DecimalFormat {

	private DecimalFormat dc = new DecimalFormat("0.0", //$NON-NLS-1$
			new DecimalFormatSymbols(Locale.US));

	public Formatter() {

	}

	/**
	 * gibt eine Zeichenkette mit <i>n </i> Nullen zur?ck
	 * 
	 * @param n
	 *            Anzahl der zu erzeugenden Nullen
	 */
	private String createzeros(int n) {
		String str = ""; //$NON-NLS-1$
		for (int i = 0; i < n; i++)
			str += "0"; //$NON-NLS-1$

		return str;
	}

	/**
	 * gibt eine Zeichenkette mit <i>n </i> "#"-Zeichen zur?ck
	 * 
	 * @param n
	 *            Anzahl der zu erzeugenden "#"
	 */
	private String createraute(int n) {
		String str = ""; //$NON-NLS-1$
		for (int i = 0; i < n; i++)
			str += "#"; //$NON-NLS-1$

		return str;
	}

	/**
	 * Rundet auf eine ganze Zahl
	 * 
	 * @param x
	 *            ein double Wert
	 */
	private String entire(double x) {
		dc.applyPattern("0"); //$NON-NLS-1$
		return dc.format(x);
	}

	/**
	 * Rundet auf eine ganze Zahl mit n "signifikanten" Stellen
	 * 
	 * Beispiel:formatentire(2.917E121,2)=2.9E121
	 * 
	 * @param x
	 *            ein double Wert
	 * @param n
	 *            Anzahl der "signifikanten" stellen n>=1
	 */
	private double formatentire(double x, int n) {

		if (n < 1)
			return Double.parseDouble(entire(x));

		dc.applyPattern(createzeros(n) + "E0"); //$NON-NLS-1$
		String temp = dc.format(x);

		return Double.valueOf(temp).doubleValue();

	}

	/**
	 * Rundet die Dezimalzahl <CODE>x</CODE> auf <CODE>n</CODE> Stellen,
	 * wobei Nullen am Ende "gezeigt" werden
	 * 
	 * @param x
	 *            eine Dezimalzahl
	 * @param places
	 *            Anzahl der Nachkommstellen >=0
	 */
	public String formatplaces(double x, int places) {
		if (places < 0)
			throw new IllegalArgumentException(""); //$NON-NLS-1$
		if (places == 0)
			return entire(x);
		else {
			dc.applyPattern("0." + createzeros(places)); //$NON-NLS-1$
			return dc.format(x);
		}
	}

	/**
	 * Rundet die Dezimalzahl <CODE>x</CODE> auf <CODE>n</CODE> Stellen,
	 * wobei Nullen am Ende nicht "gezeigt" werden
	 * 
	 * @param x
	 *            eine Dezimalzahl
	 * @param places
	 *            Anzahl der Nachkommstellen >=0
	 */
	public String formatplacesnozeros(double x, int places) {
		if (places == 0)
			return entire(x);
		else {
			dc.applyPattern("0." + createraute(places)); //$NON-NLS-1$
			return dc.format(x);
		}
	}

	private int getExp(double x) {
		dc.applyPattern("0E0"); //$NON-NLS-1$
		String temp = dc.format(x);
		int pos = temp.indexOf("E"); //$NON-NLS-1$
		return Integer.parseInt(temp.substring(pos + 1));
	}

	/**
	 * Rundet die Dezimalzahl soweit es die "Rundungsentfernung" <CODE>distance
	 * </CODE> erlaubt.
	 * 
	 * @param x
	 *            eine Dezimalzahl
	 * @param distance
	 *            max. erlaubte Differenz der gerundeten Zahl zur Zahl <CODE>x
	 *            </CODE> (dsitance>=0)
	 */
	private String formatdistance(double x, double distance) {

		if (distance < 0)
			throw new IllegalArgumentException(""); //$NON-NLS-1$

		if (distance == 0)
			return String.valueOf(x);

		if (distance > 1) {
			int i = 0;
			double temp = 0.5;
			while (temp < distance) {
				temp *= 10;
				i++;

			}
			int n = getExp(x) - i;

			return String.valueOf(formatentire(x, n));

		} else {

			int i = 0;
			double temp = 0.5;
			while (temp > distance) {
				temp /= 10;
				i++;
			}

			return formatplaces(x, i);

		}
	}

	/**
	 * Rundet die Dezimalzahl soweit es die "Rundungsentfernung" <CODE>distance
	 * </CODE> erlaubt.
	 * 
	 * @param x
	 *            eine Dezimalzahl
	 * @param distance
	 *            max. erlaubte Differenz der gerundeten Zahl zur Zahl <CODE>x
	 *            </CODE> (dsitance>=0)
	 */
	public String format(double x, double distance) {
		return String.valueOf(Double.parseDouble(formatdistance(x, distance)));

	}
}