package uibk.mtk.util;

import java.util.NoSuchElementException;

import uibk.mtk.lang.Messages;

/**
 * Eine Basisklasse f?r Iteratoren ?ber Intervalle, Linien usw.
 * 
 * 
 * <p>
 * Copyright: Institut f?r Technische Mathematik, Geometrie und Bauinformatik,
 * 2004
 * </p>
 * 
 * @author Markus Unterweger (markus.unterweger@aon.at)
 */
public abstract class ObjectIterator {

	/** untere Grenze des Intervalls */
	double a;

	/** obere Grenze des Intervalls */
	double b;

	/** Anzahl der Iterationschritte */
	int n;

	/** Schrittweite */
	protected double step;

	/** Z?hlvariable */
	protected int c;

	/** Name des Resourcebundles */
	static final String BUNDLE_NAME = "uibk.mtk.util.messages";//$NON-NLS-1$

	/**
	 * Erzeugt einen Iterator f?r <CODE>n</CODE> Werte zwischen 0 und 1
	 * 
	 * @param n
	 *            n>=2
	 */
	public ObjectIterator(int n) {
		this(0, 1, n);
	}

	/**
	 * Erzeugt einen Iterator f?r <CODE>n</CODE> Werte zwischen <CODE>a
	 * </CODE> und <CODE>b</CODE>
	 * 
	 * @param n
	 *            Anzahl der Iterationsschritte n>=2
	 * 
	 *  
	 */
	protected ObjectIterator(double a, double b, int n) {
		if (n < 2)
			throw new IllegalArgumentException(Messages.getString(BUNDLE_NAME,
					"ObjectIterator.0")); //$NON-NLS-1$
		//if (b - a < 0) throw new IllegalArgumentException("a-b<0");

		this.n = n;

		this.a = a;
		this.b = b;

		step = (b - a) / (n - 1);
	}

	/**
	 * Setzt die Iteration auf ihren urspr?nglichen Zustand zur?ck (d.h. auf den
	 * Zustand der ?quivalent ist zum Zustand nach Erzeugen eines neuen
	 * Iterators mit dem selben Konstruktor
	 */
	public void reset() {
		c = 0;
	}

	/**
	 * Gibt <CODE>true</CODE> falls die Iteration weitere Objekte hat.
	 */
	public boolean hasNext() {
		if (c < n) {
			return true;
		}
		return false;
	}

	/**
	 * Gibt die aktuelle Position der Iteration, d.h. die Anzahl der Aufrufe von
	 * <CODE>next()</CODE> zur?ck
	 */
	public int getPos() {
		return c;
	}

	/**
	 * Gibt <CODE>n</CODE> die Anzahl der Objekte die in der Iteration
	 * durchlaufen werden zur?ck.
	 */
	public int numberOfElements() {
		return n;
	}

	/**
	 * Gibt das Objekt der Iteration f?r <CODE>t</CODE>, wobei t zwischen
	 * <CODE>a</CODE> und <CODE>b</CODE> liegt zur?ck
	 *  
	 */
	protected abstract Object getObjectAt(double t);

	/**
	 * Gibt das n?chste Objekt der Iteration zur?ck
	 * 
	 * @exception NoSuchElementException
	 *                Falls es keine weiteren Objekte mehr gibt
	 */
	public Object next() throws NoSuchElementException {

		if (c >= n)
			throw new NoSuchElementException();
		double t = getT();
		c++;
		return getObjectAt(t);

	}

	/**
	 * Gibt die aktuelle Position der Interpolation zw. <CODE>a</CODE> und
	 * <CODE>b</CODE> zur?ck
	 */
	public double getT() {

		double temp = a + c * step;
		if (Math.abs(temp - b) < step / 2)
			return b;
		else
			return temp;
	}

}